# LIBTBX_SET_DISPATCHER_NAME dials.refine_bravais_settings
# LIBTBX_SET_DISPATCHER_NAME dials.rbs

"""Refinement of Bravais settings consistent with the primitive unit cell.

This program takes as input the output of dials.index, i.e. indexed.expt
and indexed.refl files. Full refinement of the crystal and experimental
geometry parameters will be performed (by default) in all Bravais settings
that are consistent with the input primitive unit cell. A table is printed
containing various information for each potential Bravais setting, including
the metric fit (a measure of the deviation from the triclinic cell),
the root-mean-square-deviations (rmsd), in mm, between the observed and
predicted spot centroids, the refined unit cell parameters in each Bravais
setting, and the change of basis operator to transform from the triclinic cell
to each Bravais setting.

The program also generates a .expt file for each Bravais setting, e.g.
bravais_setting_1.expt, which is equivalent to the input indexed.expt, but
with the crystal model refined in the chosen Bravais setting. These
bravais_setting_*.expt files are suitable as input to dials.refine or
dials.integrate, although the indexed.refl file will need to be re-indexed
using dials.reindex if the change of basis operator (cb_op) for the chosen
Bravais setting is not the identity operator (a,b,c).

Examples::

  dials.refine_bravais_settings indexed.expt indexed.refl

  dials.refine_bravais_settings indexed.expt indexed.refl nproc=4
"""


from __future__ import annotations

import collections
import json
import logging
import os
import sys

import libtbx.phil
from cctbx import sgtbx
from cctbx.sgtbx import bravais_types
from dxtbx.model import ExperimentList

import dials.util
from dials.algorithms.indexing.bravais_settings import (
    refined_settings_from_refined_triclinic,
)
from dials.array_family import flex
from dials.util import log
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.refine_bravais_settings")

phil_scope = libtbx.phil.parse(
    """
include scope dials.algorithms.indexing.bravais_settings.phil_scope

crystal_id = None
  .type = int(value_min=0)

output {
  directory = "."
    .type = path
  log = dials.refine_bravais_settings.log
    .type = path
  prefix = None
    .type = str
}

""",
    process_includes=True,
)


def bravais_lattice_to_space_groups(chiral_only=True):
    bravais_lattice_to_sg = collections.defaultdict(list)
    for sgn in range(230):
        sg = sgtbx.space_group_info(number=sgn + 1).group()
        if (not chiral_only) or (sg.is_chiral()):
            bravais_lattice = bravais_types.bravais_lattice(group=sg)
            bravais_lattice_to_sg[str(bravais_lattice)].append(sg)
    bravais_lattice_to_sg["mI"] = [sgtbx.space_group_info("I2").group()]
    return bravais_lattice_to_sg


def bravais_lattice_to_space_group_table(bravais_settings=None, chiral_only=True):
    bravais_lattice_to_sg = bravais_lattice_to_space_groups(chiral_only=chiral_only)
    logger.info("Chiral space groups corresponding to each Bravais lattice:")
    for bravais_lattice, space_groups in bravais_lattice_to_sg.items():
        if bravais_settings is not None and bravais_lattice not in bravais_settings:
            continue
        logger.info(
            ": ".join(
                [
                    bravais_lattice,
                    " ".join([short_space_group_name(sg) for sg in space_groups]),
                ]
            )
        )


def short_space_group_name(space_group):
    sgt = space_group.type()
    symbol = sgt.lookup_symbol()
    if sgt.number() > 1:
        symbol = symbol.replace(" 1", "")
    return symbol.replace(" ", "")


def eliminate_sys_absent(experiments, reflections):
    if experiments[0].crystal.get_space_group().n_ltr() > 1:
        effective_group = (
            experiments[0]
            .crystal.get_space_group()
            .build_derived_reflection_intensity_group(anomalous_flag=True)
        )
        sys_absent_flags = effective_group.is_sys_absent(reflections["miller_index"])
        reflections = reflections.select(~sys_absent_flags)
    return reflections


def map_to_primitive(experiments, reflections):
    """Map experiments and reflections to primitive setting."""
    cb_op_to_primitive = (
        experiments[0]
        .crystal.get_space_group()
        .info()
        .change_of_basis_op_to_primitive_setting()
    )
    if not cb_op_to_primitive.is_identity_op():
        logger.info(
            f"Reindexing to primitive setting with cb_op = {cb_op_to_primitive.as_abc()}"
        )
        experiments[0].crystal.update(
            experiments[0].crystal.change_basis(cb_op_to_primitive)
        )
        reflections["miller_index"] = cb_op_to_primitive.apply(
            reflections["miller_index"]
        )
    return cb_op_to_primitive


def select_datasets_on_crystal_id(experiments, reflections, crystal_id):
    """Select experiments and reflections with the given crystal id"""
    assert crystal_id < len(experiments.crystals())
    experiment_ids = experiments.where(crystal=experiments.crystals()[crystal_id])

    experiments = ExperimentList([experiments[i] for i in experiment_ids])
    refl_selections = [reflections["id"] == i for i in experiment_ids]
    reflections["id"] = flex.int(len(reflections), -1)
    for i, sel in enumerate(refl_selections):
        reflections["id"].set_selected(sel, i)
    reflections = reflections.select(reflections["id"] > -1)
    return experiments, reflections


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.refine_bravais_settings indexed.expt indexed.refl [options]"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging
    log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )
    if len(reflections) == 0 or len(experiments) == 0:
        parser.print_help()
        return

    assert len(reflections) == 1
    reflections = reflections[0]

    # Reduce what we pickle and send to workers by removing unused data
    if "shoebox" in reflections:
        del reflections["shoebox"]

    if len(experiments) == 0:
        parser.print_help()
        return
    if len(experiments.crystals()) > 1:
        if params.crystal_id is not None:
            experiments, reflections = select_datasets_on_crystal_id(
                experiments, reflections, params.crystal_id
            )
        else:
            sys.exit(
                "Only one crystal can be processed at a time: set crystal_id to choose "
                "experiment."
            )

    reflections = eliminate_sys_absent(experiments, reflections)
    cb_op_to_primitive = map_to_primitive(experiments, reflections)

    refined_settings = refined_settings_from_refined_triclinic(
        experiments,
        reflections,
        params,
        cb_op_to_primitive=cb_op_to_primitive,
    )
    possible_bravais_settings = {solution["bravais"] for solution in refined_settings}
    bravais_lattice_to_space_group_table(possible_bravais_settings)
    logger.info(refined_settings)

    prefix = params.output.prefix
    if prefix is None:
        prefix = ""
    summary_file = f"{prefix}bravais_summary.json"
    logger.info("Saving summary as %s", summary_file)
    with open(os.path.join(params.output.directory, summary_file), "w") as fh:
        json.dump(refined_settings.as_dict(), fh)

    for subgroup in refined_settings:
        expts = subgroup.refined_experiments
        soln = int(subgroup.setting_number)
        bs_json = "%sbravais_setting_%i.expt" % (prefix, soln)
        logger.info("Saving solution %i as %s", soln, bs_json)
        expts.as_file(os.path.join(params.output.directory, bs_json))


if __name__ == "__main__":
    run()



from __future__ import annotations

import random

import pytest


def test():
    from scitbx.array_family import flex

    from dials.algorithms.image.filter import summed_area

    # Create an image
    image = flex.random_double(2000 * 2000)
    image.reshape(flex.grid(2000, 2000))

    # Calculate the summed area table
    sa = summed_area(image, (3, 3))

    # For a selection of random points, ensure that the value is the
    # sum of the area under the kernel
    for i in range(10000):
        i = random.randint(10, 1990)
        j = random.randint(10, 1990)
        v = sa[j, i]
        e = flex.sum(image[j - 3 : j + 4, i - 3 : i + 4])
        assert e == pytest.approx(v, abs=1e-7)



"""Tests for the BinnedStatistics class"""

from __future__ import annotations

from scitbx.array_family import flex

from dials.algorithms.statistics import BinnedStatistics


def test_partitions():

    vals = flex.double((1, 2, 3, 1, 2, 3, 1, 2, 3))
    bins = flex.size_t((0, 1, 3, 0, 1, 3, 0, 1, 3))
    binned_statistics = BinnedStatistics(vals, bins, 4)

    assert binned_statistics.get_values_in_bin(0).all_eq(flex.double((1, 1, 1)))
    assert binned_statistics.get_values_in_bin(1).all_eq(flex.double((2, 2, 2)))
    assert binned_statistics.get_values_in_bin(2).size() == 0
    assert binned_statistics.get_values_in_bin(3).all_eq(flex.double((3, 3, 3)))
    assert binned_statistics.bin_is_empty().all_eq(
        flex.bool((False, False, True, False))
    )
    assert binned_statistics.bin_is_sorted().all_eq(
        flex.bool((False, False, False, False))
    )


def test_median_and_iqr():

    # Even number of values, one bin
    vals1 = flex.double((1, 2, 3, 4))
    bins = flex.size_t([0] * len(vals1))
    binned_statistics = BinnedStatistics(vals1, bins, 1)
    assert binned_statistics.get_medians().all_eq(flex.double((2.5,)))
    assert binned_statistics.get_iqrs().all_eq(flex.double((1.5,)))
    # NB this differs from the calculation performed via
    # _,q1,_,q3,_=five_number_summary(vals1);q3-q1
    # which gives 2.0. However, it matches the result of R's IQR function

    # Odd number of values, one bin, and robustness test
    vals2 = flex.double((1, 2, 3, 100, 1000))
    bins = flex.size_t([0] * len(vals2))
    binned_statistics = BinnedStatistics(vals2, bins, 1)
    assert binned_statistics.get_medians().all_eq(flex.double((3.0,)))
    assert binned_statistics.get_iqrs().all_eq(flex.double((98,)))
    # NB this is the same as the calculation performed via
    # _,q1,_,q3,_=five_number_summary(vals2);q3-q1, and matches R's IQR function

    # Now combine the data and randomise order
    vals3 = vals1.concatenate(vals2)
    bins = flex.size_t([0] * len(vals1) + [1] * len(vals2))
    perm = flex.random_permutation(len(bins))
    vals3 = vals3.select(perm)
    bins = bins.select(perm)
    binned_statistics = BinnedStatistics(vals3, bins, 2)
    assert binned_statistics.get_medians().all_eq(
        flex.double(
            (
                2.5,
                3.0,
            )
        )
    )
    assert binned_statistics.get_iqrs().all_eq(
        flex.double(
            (
                1.5,
                98,
            )
        )
    )



# LIBTBX_SET_DISPATCHER_NAME dials.merge_reflection_lists


from __future__ import annotations

import sys

from libtbx.phil import parse

from dials.util import show_mail_handle_errors
from dials.util.command_line import Command
from dials.util.options import ArgumentParser

# Create the help message
help_message = """

  This program allows you to combine reflection lists. Note that merge in this
  case refers to merging lists rather than merging equivalent reflections.

  Reflection tables can either be extended (i.e. given two reflection tables,
  "A" and "B", the result would have length = len("A") + len("B")). Or values
  in one table can be overwritten with values from the other table. Any columns
  present in "A" but not present in "B" (or vice versa) will be preserved.
"""

# Create the phil parameters
phil_scope = parse(
    """

  output = merged.refl
    .type = str
    .help = "The output file"

  method = *update extend
    .type = choice
    .help = "The method of merging"
"""
)


class Script:
    """A class to encapsulate the script."""

    def __init__(self):
        """Initialise the script."""
        # The script usage
        usage = "usage: dials.merge_reflection_lists [options] /path/to/image/reflection/files"
        self.parser = ArgumentParser(
            epilog=help_message, usage=usage, phil=phil_scope, read_reflections=True
        )

    def run(self, args=None):
        """Run the script."""
        # Parse the command line arguments
        params, options = self.parser.parse_args(args, show_diff_phil=True)
        if len(params.input.reflections) == 0:
            self.parser.print_help()
            return
        if len(params.input.reflections) <= 1:
            sys.exit("more than 1 reflection table must be specified")
        tables = [p.data for p in params.input.reflections]

        # Get the number of rows and columns
        nrows = [t.nrows() for t in tables]
        ncols = [t.ncols() for t in tables]

        # Merge the reflection lists
        if params.method == "update":
            assert all(n == nrows[0] for n in nrows[1:])
            table = tables[0]
            for t in tables[1:]:
                table.update(t)
        elif params.method == "extend":
            assert all(n == ncols[0] for n in ncols[1:])
            table = tables[0]
            for t in tables[1:]:
                table.extend(t)
        else:
            raise RuntimeError(f"unknown method, {params.method}")

        # Write the reflections to the file
        Command.start(f"Writing {len(table)} reflections to {params.output}")
        table.as_file(params.output)
        Command.end(f"Wrote {len(table)} reflections to {params.output}")


@show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



from __future__ import annotations

from math import pi

import wx
from annlib_ext import AnnAdaptorSelfInclude
from wx.lib.agw import floatspin

import gltbx
import gltbx.gl as gl
import libtbx.phil
import wxtbx.utils
from libtbx import Auto
from scitbx.array_family import flex
from scitbx.math import minimum_covering_sphere
from wxtbx.segmentedctrl import (
    SEGBTN_HORIZONTAL,
    SegmentedRadioControl,
    SegmentedToggleControl,
)

from dials.util import wx_viewer
from dials.util.reciprocal_lattice import Render3d

phil_scope = libtbx.phil.parse(
    """
include scope dials.util.reciprocal_lattice.phil_scope

show_rotation_axis = False
  .type = bool
show_beam_vector = False
  .type = bool
show_reciprocal_cell = False
  .type = bool
label_nearest_point = False
  .type = bool
marker_size = Auto
  .type = int(value_min=1)
autospin = False
  .type = bool
model_view_matrix = None
  .type = floats(size=16)
max_experiments = 15
  .type = int(value_min=0)
""",
    process_includes=True,
)


# WX3 - WX4 compatibility
def _rewrite_event(unbound):
    """Decorator to intercept the event and add missing instance methods"""

    def _wrapp(self, event):
        event.GetPositionTuple = event.GetPosition
        return unbound(self, event)

    return _wrapp


# HACK: Monkeypatch wxtbx so that we don't use old interfaces
wxtbx.segmentedctrl.SegmentedControl.HitTest = _rewrite_event(
    wxtbx.segmentedctrl.SegmentedControl.HitTest
)
wxtbx.segmentedctrl.SegmentedControl.OnMotion = _rewrite_event(
    wxtbx.segmentedctrl.SegmentedControl.OnMotion
)


class ReciprocalLatticeViewer(wx.Frame, Render3d):
    def __init__(self, parent, id, title, size, settings=None, *args, **kwds):
        wx.Frame.__init__(self, parent, id, title, size=size, *args, **kwds)
        Render3d.__init__(self, settings=settings)
        self.parent = self.GetParent()
        self.statusbar = self.CreateStatusBar()
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)

        self.create_settings_panel()
        self.sizer.Add(self.settings_panel, 0, wx.EXPAND)
        self.create_viewer_panel()
        self.sizer.Add(self.viewer, 1, wx.EXPAND | wx.ALL)
        self.SetSizerAndFit(self.sizer)
        self.SetMinSize(self.settings_panel.GetSize())
        self.Bind(wx.EVT_CLOSE, self.OnClose, self)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy, self)
        self.Bind(wx.EVT_ACTIVATE, self.OnActive)
        self.viewer.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.viewer.SetFocus()

    def load_models(self, experiments, reflections):
        Render3d.load_models(self, experiments, reflections)
        if self.settings.beam_centre is not None:
            self.settings_panel.beam_fast_ctrl.SetValue(self.settings.beam_centre[0])
            self.settings_panel.beam_slow_ctrl.SetValue(self.settings.beam_centre[1])
        if self.settings.marker_size is Auto:
            max_radius = max(self.reflections["rlp"].norms())
            volume = 4 / 3 * pi * max_radius**3
            density = len(self.reflections) / volume
            # Set marker size to between 5 and 50 depending on density, where
            # 1000 < density < 20000 ==> 50 < marker_size < 5
            marker_size = (-45 / 19000) * density + (5 + 900 / 19)
            marker_size = max(marker_size, 5)
            marker_size = min(marker_size, 50)
            self.settings.marker_size = marker_size
        self.settings_panel.marker_size_ctrl.SetValue(self.settings.marker_size)
        self.settings_panel.add_experiments_buttons()

    def OnActive(self, event):
        if self.IsShown() and type(self.viewer).__name__ != "_wxPyDeadObject":
            self.viewer.Refresh()

    def OnClose(self, event):
        self.Unbind(wx.EVT_ACTIVATE)
        self.Destroy()
        event.Skip()

    def OnDestroy(self, event):
        if self.parent is not None:
            self.parent.viewer = None
        event.Skip()

    def OnKeyDown(self, event):
        key = event.GetUnicodeKey()
        if key == wx.WXK_NONE:
            key = event.GetKeyCode()
        dxs = {wx.WXK_LEFT: -1, wx.WXK_RIGHT: +1, wx.WXK_UP: 0, wx.WXK_DOWN: 0}
        dys = {wx.WXK_LEFT: 0, wx.WXK_RIGHT: 0, wx.WXK_UP: +1, wx.WXK_DOWN: -1}

        if key in dxs:
            dx = dxs[key]
            dy = dys[key]
            if event.ShiftDown():
                scale = 0.1
            else:
                scale = 1.0
            self.do_Step(dx, dy, scale)

    def do_Step(self, dx, dy, scale):
        v = self.viewer
        rc = v.rotation_center
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gltbx.util.rotate_object_about_eye_x_and_y(
            scale, rc[0], rc[1], rc[2], dx, dy, 0, 0
        )
        v.OnRedraw()

    def create_viewer_panel(self):
        if self.settings.black_background:
            background_rgb = (0, 0, 0)
        else:
            background_rgb = (255, 255, 255)
        self.viewer = RLVWindow(
            settings=self.settings,
            parent=self,
            size=(800, 600),
            background_rgb=background_rgb,
        )

    def create_settings_panel(self):
        self.settings_panel = SettingsWindow(self, -1, style=wx.RAISED_BORDER)

    def set_points(self):
        Render3d.set_points(self)
        self.settings_panel.d_min_ctrl.SetValue(self.settings.d_min)
        self.settings_panel.z_min_ctrl.SetValue(self.settings.z_min)
        self.settings_panel.z_max_ctrl.SetValue(self.settings.z_max)
        self.settings_panel.n_min_ctrl.SetValue(self.settings.n_min)
        self.settings_panel.n_max_ctrl.SetValue(self.settings.n_max)
        if self.settings.partiality_min is not None:
            self.settings_panel.partiality_min_ctrl.SetValue(
                self.settings.partiality_min
            )
        if self.settings.partiality_max is not None:
            self.settings_panel.partiality_max_ctrl.SetValue(
                self.settings.partiality_max
            )

    def update_settings(self, *args, **kwds):
        self.set_beam_centre(self.settings.beam_centre_panel, self.settings.beam_centre)
        self.map_points_to_reciprocal_space()
        self.set_points()
        self.viewer.update_settings(*args, **kwds)

    def update_statusbar(self):
        model_view_matrix = gltbx.util.get_gl_modelview_matrix()
        txt = (
            "Model view matrix: "
            + "["
            + ", ".join("%.4f" % m for m in model_view_matrix)
            + "]"
        )
        self.statusbar.SetStatusText(txt)


class SettingsWindow(wxtbx.utils.SettingsPanel):
    def __init__(self, *args, **kwds):
        wxtbx.utils.SettingsPanel.__init__(self, *args, **kwds)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def OnChar(self, event):
        self.GetParent().viewer.OnChar(event)

    def add_controls(self):
        # d_min control

        self.d_min_ctrl = floatspin.FloatSpin(
            parent=self, increment=0.05, min_val=0, digits=2
        )
        self.d_min_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.d_min_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "High resolution:")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.d_min_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.d_min_ctrl)

        self.z_min_ctrl = floatspin.FloatSpin(
            parent=self, increment=1, min_val=0, digits=0
        )
        self.z_min_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.z_min_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Min Z")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.z_min_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.z_min_ctrl)

        self.z_max_ctrl = floatspin.FloatSpin(
            parent=self, increment=1, min_val=0, digits=0
        )
        self.z_max_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.z_max_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Max Z")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.z_max_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.z_max_ctrl)

        # Control for spot size (utility depends on n_signal column in reflection
        # file - will be ignored if not in file

        self.n_min_ctrl = floatspin.FloatSpin(
            parent=self, increment=1, min_val=0, digits=0
        )
        self.n_min_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.n_min_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Min Pixels")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.n_min_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.n_min_ctrl)

        self.n_max_ctrl = floatspin.FloatSpin(
            parent=self, increment=1, min_val=0, digits=0
        )
        self.n_max_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.n_max_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Max Pixels")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.n_max_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.n_max_ctrl)

        # end new control

        self.partiality_min_ctrl = floatspin.FloatSpin(
            parent=self, increment=0.01, digits=3, min_val=0, max_val=1
        )
        self.partiality_min_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.partiality_min_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Min partiality")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.partiality_min_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(
            floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.partiality_min_ctrl
        )

        self.partiality_max_ctrl = floatspin.FloatSpin(
            parent=self, increment=0.01, digits=3, min_val=0, max_val=1
        )
        self.partiality_max_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.partiality_max_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Max partiality")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.partiality_max_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(
            floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.partiality_max_ctrl
        )

        ctrls = self.create_controls(
            setting="show_rotation_axis", label="Show rotation axis"
        )
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)
        ctrls = self.create_controls(
            setting="show_beam_vector", label="Show beam vector"
        )
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)
        ctrls = self.create_controls(
            setting="show_reciprocal_cell", label="Show reciprocal cell"
        )
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)
        ctrls = self.create_controls(
            setting="label_nearest_point", label="Label nearest point"
        )
        self.panel_sizer.Add(ctrls[0], 0, wx.ALL, 5)
        self.reverse_phi_ctrl = self.create_controls(
            setting="reverse_phi", label="Invert rotation axis"
        )[0]
        self.panel_sizer.Add(self.reverse_phi_ctrl, 0, wx.ALL, 5)

        self.Bind(wx.EVT_CHECKBOX, self.OnChangeSettings, self.reverse_phi_ctrl)

        self.crystal_frame_tooltip = wx.ToolTip(
            "Show the reciprocal lattice(s) in the crystal rather than the laboratory frame"
        )
        self.crystal_frame_ctrl = self.create_controls(
            setting="crystal_frame", label="Show in crystal frame"
        )[0]
        self.crystal_frame_ctrl.SetToolTip(self.crystal_frame_tooltip)
        self.panel_sizer.Add(self.crystal_frame_ctrl, 0, wx.ALL, 5)

        self.Bind(wx.EVT_CHECKBOX, self.OnChangeSettings, self.crystal_frame_ctrl)

        self.beam_panel_ctrl = floatspin.FloatSpin(
            parent=self, min_val=0, increment=1, digits=0
        )
        self.beam_panel_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.beam_panel_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Beam centre panel")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.beam_panel_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.beam_panel_ctrl)

        self.beam_fast_ctrl = floatspin.FloatSpin(parent=self, increment=0.01, digits=2)
        self.beam_fast_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.beam_fast_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Beam centre (mm)")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.beam_fast_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.beam_fast_ctrl)

        self.beam_slow_ctrl = floatspin.FloatSpin(parent=self, increment=0.01, digits=2)
        self.beam_slow_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.beam_slow_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box.Add(self.beam_slow_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.beam_slow_ctrl)

        self.marker_size_ctrl = floatspin.FloatSpin(
            parent=self, increment=1, digits=0, min_val=1
        )
        self.marker_size_ctrl.Bind(wx.EVT_SET_FOCUS, lambda evt: None)
        if wx.VERSION >= (2, 9):  # XXX FloatSpin bug in 2.9.2/wxOSX_Cocoa
            self.marker_size_ctrl.SetBackgroundColour(self.GetBackgroundColour())
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Marker size:")
        box.Add(label, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        box.Add(self.marker_size_ctrl, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(floatspin.EVT_FLOATSPIN, self.OnChangeSettings, self.marker_size_ctrl)

        self.btn = SegmentedRadioControl(self, style=SEGBTN_HORIZONTAL)
        self.btn.AddSegment("all")
        self.btn.AddSegment("indexed")
        self.btn.AddSegment("unindexed")
        self.btn.AddSegment("integrated")
        self.btn.SetSelection(
            ["all", "indexed", "unindexed", "integrated"].index(self.settings.display)
        )
        self.Bind(wx.EVT_RADIOBUTTON, self.OnChangeSettings, self.btn)
        self.GetSizer().Add(self.btn, 0, wx.ALL, 5)

        self.outlier_btn = SegmentedRadioControl(self, style=SEGBTN_HORIZONTAL)
        self.outlier_btn.AddSegment("all")
        self.outlier_btn.AddSegment("inliers")
        self.outlier_btn.AddSegment("outliers")
        self.outlier_btn.SetSelection(
            [None, "inliers", "outliers"].index(self.settings.outlier_display)
        )
        self.Bind(wx.EVT_RADIOBUTTON, self.OnChangeSettings, self.outlier_btn)
        self.GetSizer().Add(self.outlier_btn, 0, wx.ALL, 5)

    def add_value_widgets(self, sizer):
        sizer.Add(
            wx.StaticText(self.panel, -1, "Value:"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.value_info = wx.TextCtrl(
            self.panel, -1, size=(80, -1), style=wx.TE_READONLY
        )
        sizer.Add(self.value_info, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)

    def add_experiments_buttons(self):
        n = flex.max(self.parent.reflections_input["id"])
        if n <= 0 or n > self.settings.max_experiments:
            self.expt_btn = None
            return

        box = wx.BoxSizer(wx.VERTICAL)
        self.panel_sizer.Add(box)
        label = wx.StaticText(self, -1, "Experiment ids:")
        box.Add(label, 0, wx.ALL, 5)

        self.expt_btn = SegmentedToggleControl(self, style=SEGBTN_HORIZONTAL)
        for i in range(-1, n + 1):
            self.expt_btn.AddSegment(str(i))
            if (
                self.settings.experiment_ids is not None
                and i in self.settings.experiment_ids
            ):
                self.expt_btn.SetValue(i + 1, True)

        self.expt_btn.Realize()
        self.Bind(wx.EVT_TOGGLEBUTTON, self.OnChangeSettings, self.expt_btn)
        box.Add(self.expt_btn, 0, wx.ALL, 5)

    def OnChangeSettings(self, event):
        self.settings.d_min = self.d_min_ctrl.GetValue()
        self.settings.z_min = self.z_min_ctrl.GetValue()
        self.settings.z_max = self.z_max_ctrl.GetValue()
        self.settings.n_min = int(self.n_min_ctrl.GetValue())
        self.settings.n_max = int(self.n_max_ctrl.GetValue())
        self.settings.partiality_min = self.partiality_min_ctrl.GetValue()
        self.settings.partiality_max = self.partiality_max_ctrl.GetValue()

        old_beam_panel = self.settings.beam_centre_panel
        old_beam_centre = self.settings.beam_centre
        self.settings.beam_centre_panel = self.beam_panel_ctrl.GetValue()
        self.settings.beam_centre = (
            self.beam_fast_ctrl.GetValue(),
            self.beam_slow_ctrl.GetValue(),
        )
        self.settings.reverse_phi = self.reverse_phi_ctrl.GetValue()
        self.settings.crystal_frame = self.crystal_frame_ctrl.GetValue()
        self.settings.marker_size = self.marker_size_ctrl.GetValue()
        for i, display in enumerate(("all", "indexed", "unindexed", "integrated")):
            if self.btn.values[i]:
                self.settings.display = display
                break

        for i, display in enumerate(("all", "inliers", "outliers")):
            if self.outlier_btn.values[i]:
                self.settings.outlier_display = display
                break

        if self.expt_btn is not None:
            expt_ids = []
            for i in range(len(self.expt_btn.segments)):
                if self.expt_btn.GetValue(i):
                    expt_ids.append(i - 1)
            self.settings.experiment_ids = expt_ids

        try:
            self.parent.update_settings()
        except ValueError:  # Handle beam centre changes, which could fail
            self.settings.beam_centre_panel = old_beam_panel
            self.settings.beam_centre = old_beam_centre
            self.beam_panel_ctrl.SetValue(old_beam_panel)
            self.beam_fast_ctrl.SetValue(old_beam_centre[0])
            self.beam_slow_ctrl.SetValue(old_beam_centre[1])


class RLVWindow(wx_viewer.show_points_and_lines_mixin):
    def __init__(self, settings, *args, **kwds):
        super().__init__(*args, **kwds)
        self.settings = settings
        self.points = flex.vec3_double()
        self.colors = None
        self.palette = None
        self.rotation_axis = None
        self.beam_vector = None
        self.recip_latt_vectors = None
        self.recip_crystal_vectors = None
        self.flag_show_minimum_covering_sphere = False
        self.minimum_covering_sphere = None
        self.field_of_view_y = 0.001
        if self.settings.autospin:
            self.autospin_allowed = True
            self.yspin = 1
            self.xspin = 1
            self.autospin = True

    def set_points(self, points):
        self.points = points
        self.points_display_list = None
        if self.minimum_covering_sphere is None:
            self.update_minimum_covering_sphere()

    def set_points_data(self, reflections):
        dstar = reflections["rlp"].norms()
        dstar.set_selected(dstar == 0, 1e-8)
        self.points_data = {
            "panel": reflections["panel"],
            "id": reflections["id"],
            "xyz": reflections["xyzobs.px.value"],
            "d_spacing": 1 / dstar,
        }
        if "miller_index" in reflections:
            self.points_data["miller_index"] = reflections["miller_index"]

    def set_colors(self, colors):
        assert len(colors) == len(self.points)
        self.colors = colors

    def set_palette(self, palette):
        self.palette = palette

    def draw_points(self):
        if self.points_display_list is None:
            self.points_display_list = gltbx.gl_managed.display_list()
            self.points_display_list.compile()
            gl.glLineWidth(1)
            if self.colors is None:
                self.colors = flex.vec3_double(len(self.points), (1, 1, 1))
            for point, color in zip(self.points, self.colors):
                self.draw_cross_at(point, color=color)
            self.points_display_list.end()
        self.points_display_list.call()

    def set_rotation_axis(self, axis):
        self.rotation_axis = axis

    def set_beam_vector(self, beam):
        self.beam_vector = beam

    def set_reciprocal_lattice_vectors(self, vectors_per_crystal):
        self.recip_latt_vectors = vectors_per_crystal

    def set_reciprocal_crystal_vectors(self, vectors_per_crystal):
        self.recip_crystal_vectors = vectors_per_crystal

    # --- user input and settings
    def update_settings(self):
        self.points_display_list = None
        self.Refresh()

    def update_minimum_covering_sphere(self):
        n_points = min(1000, self.points.size())
        isel = flex.random_permutation(self.points.size())[:n_points]
        self.minimum_covering_sphere = minimum_covering_sphere(self.points.select(isel))

    def draw_cross_at(self, xyz, color=(1, 1, 1), f=None):
        (x, y, z) = xyz
        if f is None:
            f = 0.01 * self.settings.marker_size
        wx_viewer.show_points_and_lines_mixin.draw_cross_at(
            self, (x, y, z), color=color, f=f
        )

    def DrawGL(self):
        wx_viewer.show_points_and_lines_mixin.DrawGL(self)
        if self.rotation_axis is not None and self.settings.show_rotation_axis:
            self.draw_axis(self.rotation_axis, "phi")
        if self.beam_vector is not None and self.settings.show_beam_vector:
            self.draw_axis(self.beam_vector, "beam")

        if self.settings.show_reciprocal_cell:
            # if we don't have one sort of vector we don't have the other either
            vectors = self.recip_latt_vectors
            if self.settings.crystal_frame:
                vectors = self.recip_crystal_vectors

            if vectors:
                for i, axes in enumerate(vectors):
                    if self.settings.experiment_ids:
                        if i not in self.settings.experiment_ids:
                            continue
                    j = (i + 1) % self.palette.size()
                    color = self.palette[j]
                    self.draw_cell(axes, color)

        if self.settings.label_nearest_point:
            self.label_nearest_point()

        self.GetParent().update_statusbar()

    def draw_axis(self, axis, label):
        if self.minimum_covering_sphere is None:
            self.update_minimum_covering_sphere()
        s = self.minimum_covering_sphere
        scale = max(max(s.box_max()), abs(min(s.box_min())))
        gltbx.fonts.ucs_bitmap_8x13.setup_call_lists()
        gl.glDisable(gl.GL_LIGHTING)
        if self.settings.black_background:
            gl.glColor3f(1.0, 1.0, 1.0)
        else:
            gl.glColor3f(0.0, 0.0, 0.0)
        gl.glLineWidth(1.0)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(axis[0] * scale, axis[1] * scale, axis[2] * scale)
        gl.glEnd()
        gl.glRasterPos3f(
            0.5 + axis[0] * scale, 0.2 + axis[1] * scale, 0.2 + axis[2] * scale
        )
        gltbx.fonts.ucs_bitmap_8x13.render_string(label)
        gl.glEnable(gl.GL_LINE_STIPPLE)
        gl.glLineStipple(4, 0xAAAA)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(-axis[0] * scale, -axis[1] * scale, -axis[2] * scale)
        gl.glEnd()
        gl.glDisable(gl.GL_LINE_STIPPLE)

    def draw_cell(self, axes, color):
        astar, bstar, cstar = axes[0], axes[1], axes[2]
        gltbx.fonts.ucs_bitmap_8x13.setup_call_lists()
        gl.glDisable(gl.GL_LIGHTING)
        gl.glColor3f(*color)
        gl.glLineWidth(2.0)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(*astar.elems)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(*bstar.elems)
        gl.glVertex3f(0.0, 0.0, 0.0)
        gl.glVertex3f(*cstar.elems)
        gl.glEnd()
        gl.glRasterPos3f(*(1.01 * astar).elems)
        gltbx.fonts.ucs_bitmap_8x13.render_string("a*")
        gl.glRasterPos3f(*(1.01 * bstar).elems)
        gltbx.fonts.ucs_bitmap_8x13.render_string("b*")
        gl.glRasterPos3f(*(1.01 * cstar).elems)
        gltbx.fonts.ucs_bitmap_8x13.render_string("c*")
        gl.glEnable(gl.GL_LINE_STIPPLE)
        gl.glLineStipple(4, 0xAAAA)
        farpoint = astar + bstar + cstar
        # a* face
        gl.glBegin(gl.GL_LINE_LOOP)
        gl.glVertex3f(*farpoint.elems)
        gl.glVertex3f(*(farpoint - bstar).elems)
        gl.glVertex3f(*(farpoint - bstar - cstar).elems)
        gl.glVertex3f(*(farpoint - cstar).elems)
        gl.glEnd()
        # b* face
        gl.glBegin(gl.GL_LINE_LOOP)
        gl.glVertex3f(*farpoint.elems)
        gl.glVertex3f(*(farpoint - astar).elems)
        gl.glVertex3f(*(farpoint - astar - cstar).elems)
        gl.glVertex3f(*(farpoint - cstar).elems)
        gl.glEnd()
        # c* face
        gl.glBegin(gl.GL_LINE_LOOP)
        gl.glVertex3f(*farpoint.elems)
        gl.glVertex3f(*(farpoint - bstar).elems)
        gl.glVertex3f(*(farpoint - bstar - astar).elems)
        gl.glVertex3f(*(farpoint - astar).elems)
        gl.glEnd()
        gl.glDisable(gl.GL_LINE_STIPPLE)

    def label_nearest_point(self):
        ann = AnnAdaptorSelfInclude(self.points.as_double(), 3)
        ann.query(self.rotation_center)
        i = ann.nn[0]
        gltbx.fonts.ucs_bitmap_8x13.setup_call_lists()
        gl.glDisable(gl.GL_LIGHTING)
        gl.glColor3f(1.0, 1.0, 1.0)
        gl.glLineWidth(1.0)
        xyz = self.points_data["xyz"][i]
        exp_id = self.points_data["id"][i]
        panel = self.points_data["panel"][i]
        d_spacing = self.points_data["d_spacing"][i]
        label = (
            f"id: {exp_id}; panel: {panel}\n"
            f"xyz: {xyz[0]:.1f} {xyz[1]:.1f} {xyz[2]:.1f}\n"
            f"res: {d_spacing:.2f} Angstrom"
        )
        if "miller_index" in self.points_data and exp_id != -1:
            hkl = self.points_data["miller_index"][i]
            label += f"\nhkl: {hkl}"
        line_spacing = round(gltbx.fonts.ucs_bitmap_8x13.height())
        for j, string in enumerate(label.splitlines()):
            gl.glRasterPos3f(*self.points[i])
            gl.glBitmap(0, 0, 0.0, 0.0, line_spacing, -j * line_spacing, b" ")
            gltbx.fonts.ucs_bitmap_8x13.render_string(string)

    def rotate_view(self, x1, y1, x2, y2, shift_down=False, scale=0.1):
        super().rotate_view(x1, y1, x2, y2, shift_down=shift_down, scale=scale)

    def OnLeftUp(self, event):
        self.was_dragged = True
        super().OnLeftUp(event)

    def initialize_modelview(self, eye_vector=None, angle=None):
        super().initialize_modelview(eye_vector=eye_vector, angle=angle)
        self.rotation_center = (0, 0, 0)
        self.move_to_center_of_viewport(self.rotation_center)
        if self.settings.model_view_matrix is not None:
            gl.glLoadMatrixd(self.settings.model_view_matrix)



from __future__ import annotations


class DialsRefineConfigError(ValueError):
    pass


class DialsRefineRuntimeError(RuntimeError):
    pass


from dials.algorithms.refinement.refiner import Refiner, RefinerFactory

__all__ = [
    "DialsRefineConfigError",
    "DialsRefineRuntimeError",
    "Refiner",
    "RefinerFactory",
]



from __future__ import annotations

from scitbx.array_family import flex


class SpotMatcher:
    """Match the observed with predicted spots."""

    def __init__(self, max_separation=2):
        """
        Setup the algorithm

        :param max_separation: Max pixel dist between predicted and observed spot
        """
        # Set the algorithm parameters
        self._max_separation = max_separation

    def __call__(self, observed, predicted):
        """
        Match the observed reflections with the predicted.

        :param observed: The list of observed reflections.
        :param predicted: The list of predicted reflections.

        :returns: The list of matched reflections
        """
        from dials.array_family import flex

        # Find the nearest neighbours and distances
        nn, dist = self._find_nearest_neighbours(observed, predicted)

        # Filter the matches by distance
        index = self._filter_by_distance(nn, dist)

        # Filter out duplicates to just leave the closest pairs
        index = self._filter_duplicates(index, nn, dist)

        # Copy all of the reflection data for the matched reflections
        return flex.size_t(index), flex.size_t([nn[i] for i in index])

    def _find_nearest_neighbours(self, observed, predicted):
        """
        Find the nearest predicted spot to the observed spot.

        :param observed: The observed reflections
        :param predicted: The predicted reflections

        :returns: (nearest neighbours, distance)
        """
        # Get the predicted coordinates
        predicted_panel = predicted["panel"]
        predicted_xyz = predicted["xyzcal.px"]
        observed_panel = observed["panel"]
        observed_xyz = observed["xyzobs.px.value"]

        # Get the number of panels
        max_panel1 = flex.max(predicted_panel)
        max_panel2 = flex.max(observed_panel)
        max_panel = max([max_panel1, max_panel2])

        nn_all = flex.size_t()
        dd_all = flex.double()
        for panel in range(max_panel + 1):
            pind = predicted_panel == panel
            oind = observed_panel == panel
            pxyz = predicted_xyz.select(pind)
            oxyz = observed_xyz.select(oind)
            nn, d = self._find_nearest_neighbours_single(oxyz, pxyz)
            indices = flex.size_t(range(len(pind))).select(pind)
            indices = indices.select(flex.size_t(list(nn)))
            nn_all.extend(indices)
            dd_all.extend(d)
        return nn_all, dd_all

    def _find_nearest_neighbours_single(self, oxyz, pxyz):
        """
        Find the nearest predicted spot to the observed spot.

        :param observed: The observed reflections
        :param predicted: The predicted reflections

        :returns: (nearest neighbours, distance)
        """
        from annlib_ext import AnnAdaptor

        # Create the KD Tree
        ann = AnnAdaptor(pxyz.as_double().as_1d(), 3)

        # Query to find all the nearest neighbours
        ann.query(oxyz.as_double().as_1d())

        # Return the nearest neighbours and distances
        return ann.nn, flex.sqrt(ann.distances)

    def _filter_by_distance(self, nn, dist):
        """
        Filter the matches by distance.

        :param nn: The nearest neighbour list
        :param dist: The distances

        :returns: A reduced list of nearest neighbours
        """
        index = range(len(nn))
        return flex.int(i for i in index if dist[i] <= self._max_separation)

    def _filter_duplicates(self, index, nn, dist):
        """
        Filter the matches to remove duplicates

        :param index: The indices of valid spots
        :param nn: The nearest neighbour indices
        :param dist: The distances

        :returns: A reduced list of nearest neighbours
        """
        seen = {}
        for i in index:
            p = nn[i]
            if p in seen:
                j = seen[p]
                if dist[i] < dist[j]:
                    seen[p] = i
            else:
                seen[p] = i

        index = list(seen.values())
        return index



from __future__ import annotations

import pytest


@pytest.fixture
def observation():
    from dials.model.data import Observation

    obs = Observation()
    obs.panel = 1
    obs.centroid.px.position = (100, 200, 300)
    obs.centroid.px.variance = (10, 20, 30)
    obs.centroid.px.std_err_sq = (1, 2, 3)
    obs.centroid.mm.position = (100.1, 200.1, 300.1)
    obs.centroid.mm.variance = (10.1, 20.1, 30.1)
    obs.centroid.mm.std_err_sq = (1.1, 2.1, 3.1)
    obs.intensity.observed.value = 1234
    obs.intensity.observed.variance = 42
    obs.intensity.corrected.value = 1234.1
    obs.intensity.corrected.variance = 42.1
    return obs


def test_data(observation):
    eps = 1e-7

    assert observation.panel == 1

    assert abs(observation.centroid.px.position[0] - 100) < eps
    assert abs(observation.centroid.px.position[1] - 200) < eps
    assert abs(observation.centroid.px.position[2] - 300) < eps
    assert abs(observation.centroid.px.variance[0] - 10) < eps
    assert abs(observation.centroid.px.variance[1] - 20) < eps
    assert abs(observation.centroid.px.variance[2] - 30) < eps
    assert abs(observation.centroid.px.std_err_sq[0] - 1) < eps
    assert abs(observation.centroid.px.std_err_sq[1] - 2) < eps
    assert abs(observation.centroid.px.std_err_sq[2] - 3) < eps

    assert abs(observation.centroid.mm.position[0] - 100.1) < eps
    assert abs(observation.centroid.mm.position[1] - 200.1) < eps
    assert abs(observation.centroid.mm.position[2] - 300.1) < eps
    assert abs(observation.centroid.mm.variance[0] - 10.1) < eps
    assert abs(observation.centroid.mm.variance[1] - 20.1) < eps
    assert abs(observation.centroid.mm.variance[2] - 30.1) < eps
    assert abs(observation.centroid.mm.std_err_sq[0] - 1.1) < eps
    assert abs(observation.centroid.mm.std_err_sq[1] - 2.1) < eps
    assert abs(observation.centroid.mm.std_err_sq[2] - 3.1) < eps

    assert abs(observation.intensity.observed.value - 1234) < eps
    assert abs(observation.intensity.observed.variance - 42) < eps
    assert abs(observation.intensity.corrected.value - 1234.1) < eps
    assert abs(observation.intensity.corrected.variance - 42.1) < eps


def test_equality(observation):
    from dials.model.data import Observation

    obs2 = Observation(observation)
    assert obs2 == observation
    obs2.panel += 1
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.px.position = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.px.variance = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.px.std_err_sq = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.mm.position = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.mm.variance = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.centroid.mm.std_err_sq = (0, 0, 0)
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.intensity.observed.value = 0
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.intensity.observed.variance = 0
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.intensity.corrected.value = 0
    assert obs2 != observation
    obs2 = Observation(observation)
    obs2.intensity.corrected.variance = 0
    assert obs2 != observation



"""
Test for the basis function module.
"""

from __future__ import annotations

import pytest

from dials.algorithms.scaling.basis_functions import RefinerCalculator
from dials.algorithms.scaling.model.components.scale_components import (
    SingleBScaleFactor,
    SingleScaleFactor,
)
from dials.algorithms.scaling.parameter_handler import scaling_active_parameter_manager
from dials.array_family import flex


@pytest.fixture
def small_reflection_table():
    """Generate reflection table to test the basis function."""
    reflections = flex.reflection_table()
    reflections["d"] = flex.double([2.0, 0.8, 2.0])  # don't change
    reflections["xyzobs.px.value"] = flex.vec3_double(
        [(0.0, 0.0, 0.0), (0.0, 0.0, 5.0), (0.0, 0.0, 10.0)]
    )
    reflections["id"] = flex.int(3, 0)
    return reflections


def test_RefinerCalculator(small_reflection_table):
    """Test for the RefinerCalculator class. This calculates scale factors and
    derivatives for reflections based on the model components."""

    # To test the basis function, need a scaling active parameter manager - to set
    # this up we need a components dictionary with some reflection data.

    # Let's use KB model components for simplicity - and have an extra fake 'abs'
    # component.
    rt = small_reflection_table
    components = {
        "scale": SingleScaleFactor(flex.double([1.0])),
        "decay": SingleBScaleFactor(flex.double([0.0])),
        "abs": SingleScaleFactor(flex.double([1.0])),
    }  # Create empty components.
    components["scale"].data = {"id": rt["id"]}
    components["decay"].data = {"d": rt["d"]}
    components["abs"].data = {"id": rt["id"]}
    for component in components.values():
        component.update_reflection_data()  # Add some data to components.

    apm = scaling_active_parameter_manager(components, ["decay", "scale"])

    # First test that scale factors can be successfully updated.
    # Manually change the parameters in the apm.
    decay = components["decay"]  # Define alias
    _ = components["scale"]  # Define alias
    # Note, order of params in apm.x depends on order in scaling model components.
    new_B = 1.0
    new_S = 2.0
    apm.set_param_vals(flex.double([new_S, new_B]))
    s, d = RefinerCalculator.calculate_scales_and_derivatives(apm, 0)
    slist, dlist = RefinerCalculator._calc_component_scales_derivatives(apm, 0)
    # Now test that the inverse scale factor is correctly calculated.
    calculated_sfs = s
    assert list(calculated_sfs) == pytest.approx(
        list(new_S * flex.exp(new_B / (2.0 * flex.pow2(decay.d_values[0]))))
    )

    # Now check that the derivative matrix is correctly calculated.
    calc_derivs = d
    assert calc_derivs[0, 0] == dlist[0][0, 0] * slist[1][0]
    assert calc_derivs[1, 0] == dlist[0][1, 0] * slist[1][1]
    assert calc_derivs[2, 0] == dlist[0][2, 0] * slist[1][2]
    assert calc_derivs[0, 1] == dlist[1][0, 0] * slist[0][0]
    assert calc_derivs[1, 1] == dlist[1][1, 0] * slist[0][1]
    assert calc_derivs[2, 1] == dlist[1][2, 0] * slist[0][2]

    # Repeat the test when there is only one active parameter.
    # First reset the parameters
    components["decay"].parameters = flex.double([0.0])
    components["scale"].parameters = flex.double([1.0])
    components["abs"].parameters = flex.double([1.0])
    components["decay"].calculate_scales_and_derivatives()
    components["scale"].calculate_scales_and_derivatives()
    components["abs"].calculate_scales_and_derivatives()

    # Now generate a parameter manager for a single component.
    apm = scaling_active_parameter_manager(components, ["scale"])
    new_S = 2.0
    apm.set_param_vals(flex.double(components["scale"].n_params, new_S))
    s, d = RefinerCalculator.calculate_scales_and_derivatives(apm, 0)
    slist, dlist = RefinerCalculator._calc_component_scales_derivatives(apm, 0)
    # Test that the scales and derivatives were correctly calculated
    assert list(s) == list([new_S] * slist[0].size())
    assert d[0, 0] == dlist[0][0, 0]
    assert d[1, 0] == dlist[0][1, 0]
    assert d[2, 0] == dlist[0][2, 0]

    # Test again for two components.
    components["decay"].parameters = flex.double([0.0])
    components["scale"].parameters = flex.double([1.0])
    components["abs"].parameters = flex.double([1.0])
    components["decay"].calculate_scales_and_derivatives()
    components["scale"].calculate_scales_and_derivatives()
    components["abs"].calculate_scales_and_derivatives()

    apm = scaling_active_parameter_manager(components, ["scale", "decay"])
    _, __ = RefinerCalculator.calculate_scales_and_derivatives(apm, 0)

    # Test for no components
    apm = scaling_active_parameter_manager(components, [])
    _, d = RefinerCalculator.calculate_scales_and_derivatives(apm, 0)
    assert d.n_cols == 0 and d.n_rows == 0



from __future__ import annotations

from pathlib import Path

import procrunner
import pytest

from dxtbx.serialize import load

from dials.array_family import flex


def test_slice_sequence_and_compare_with_expected_results(dials_regression, tmp_path):
    # use the i04_weak_data for this test
    data_dir = Path(dials_regression, "refinement_test_data", "i04_weak_data")
    experiments_path = data_dir / "experiments.json"
    pickle_path = data_dir / "indexed_strong.pickle"

    assert experiments_path.is_file()
    assert pickle_path.is_file()

    result = procrunner.run(
        ["dials.slice_sequence", experiments_path, pickle_path, "image_range=1 20"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # load results
    sliced_exp = load.experiment_list(
        tmp_path / "experiments_1_20.expt", check_format=False
    )[0]
    sliced_refs = flex.reflection_table.from_file(tmp_path / "indexed_strong_1_20.refl")

    # simple test of results
    assert sliced_exp.scan.get_image_range() == (1, 20)
    assert len(sliced_refs) == 3670


def test_slice_sequence_with_first_images_missing(dials_regression, tmp_path):
    """Test slicing where scan image range does not start at 1, exercising
    a case that exposed a bug"""

    # use the i04_weak_data for this test
    data_dir = Path(dials_regression, "refinement_test_data", "i04_weak_data")
    experiments_path = data_dir / "experiments.json"

    # first slice
    result = procrunner.run(
        ["dials.slice_sequence", experiments_path, "image_range=5,20"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # second slice
    result = procrunner.run(
        ["dials.slice_sequence", "experiments_5_20.expt", "image_range=10,20"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    sliced_exp = load.experiment_list(
        tmp_path / "experiments_5_20_10_20.expt", check_format=False
    )[0]
    assert sliced_exp.scan.get_image_range() == (10, 20)
    assert sliced_exp.scan.get_array_range() == (9, 20)
    assert sliced_exp.scan.get_oscillation()[0] == pytest.approx(83.35)


def test_slice_sequence_to_degree_blocks(dials_data, tmp_path):
    """Slice data into 10 degree blocks i.e. 17 datasets"""
    expt = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True) / "scaled_30.expt"
    refl = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True) / "scaled_30.refl"
    procrunner.run(
        [
            "dials.slice_sequence",
            "block_size=10",
            "output.experiments=sliced.expt",
            "output.reflections=sliced.refl",
            expt,
            refl,
        ],
        working_directory=tmp_path,
    )

    sliced_expts = load.experiment_list(tmp_path / "sliced.expt", check_format=False)
    assert len(sliced_expts) == 17
    sliced_refl = flex.reflection_table.from_file(tmp_path / "sliced.refl")
    assert len(set(sliced_refl.experiment_identifiers().values())) == 17
    sliced_refl.assert_experiment_identifiers_are_consistent(sliced_expts)


def test_slice_sequence_with_scan_varying_crystal(dials_data, tmp_path):
    """test slicing keeps a scan-varying crystal"""

    expt = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True) / "scaled_30.expt"
    procrunner.run(
        [
            "dials.slice_sequence",
            "image_range=10,20",
            "output.experiments=sliced.expt",
            expt,
        ],
        working_directory=tmp_path,
    )
    orig = load.experiment_list(expt, check_format=False)[0]

    sliced = load.experiment_list(tmp_path / "sliced.expt", check_format=False)[0]

    assert sliced.crystal.num_scan_points == 12

    orig_UB = [
        orig.crystal.get_A_at_scan_point(i) for i in range(orig.crystal.num_scan_points)
    ]
    sliced_UB = [
        sliced.crystal.get_A_at_scan_point(i)
        for i in range(sliced.crystal.num_scan_points)
    ]

    for a, b in zip(orig_UB[9:21], sliced_UB):
        assert a == pytest.approx(b)



"""
Test extracting of reflection intensities for export.

Two main use cases supported:

- exporting after scaling, should export only the 'scale' intensities and
  variances with the inverse scale correction applied.
  dials.scale should choose the best intensities, filter outliers and apply all
  corrections except the inverse scale factor to create the
  'intensity.scale.value', so all that is needed is to apply the inverse scales
  and propagate any errors

- exporting after integration, to send for scaling outside of dials. In this
  case, one needs to decide which intensities to export, prf, sum or both.
  this can be split into two further cases:
  - export only prf or only sum
  - export both prf and sum, even if both might not be defined for all
    reflections, so that downstream programs can decide how to use both
    intensities

As well as choosing the correct intensities, further filtering may be applied
- removing ice rings
- only exporting above a minimimum I/sigma
- filter out 'bad' values such as negative variances, negative scale factors

Lastly, partials should be correctly handled - being combined if appropriate,
scaled and filtered below a given value where the values become unreliable.
"""

from __future__ import annotations

from unittest import mock

import pytest

from cctbx import miller
from dxtbx.model import Crystal, Experiment, ExperimentList
from dxtbx.model.experiment_list import ExperimentListFactory

from dials.array_family import flex
from dials.util.filter_reflections import (
    AllSumPrfScaleIntensityReducer,
    FilterForExportAlgorithm,
    FilteringReductionMethods,
    NoProfilesException,
    PrfIntensityReducer,
    ScaleIntensityReducer,
    SumAndPrfIntensityReducer,
    SumIntensityReducer,
    _sum_prf_partials,
    _sum_scale_partials,
    _sum_sum_partials,
    filter_reflection_table,
    filtered_arrays_from_experiments_reflections,
    sum_partial_reflections,
)


def generate_simple_table():
    """Generate a simple table for testing export function."""
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.prf.variance"] = flex.double([1.0, 1.0, 1.0])
    r["partiality"] = flex.double([1.0, 1.0, 0.5])
    r["partial_id"] = flex.int([0, 1, 2])
    r["id"] = flex.int(3, 0)
    r.set_flags(flex.bool([True, True, True]), r.flags.integrated_prf)
    r.set_flags(flex.bool([True, False, False]), r.flags.in_powder_ring)
    return r


def generate_integrated_test_reflections():
    """Generate an example reflection table that would be generated by
    dials.integrate or dials.scale. This is a simple example with no partials or
    corrections, to test the correct selection of intensities."""
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
    r["intensity.sum.value"] = flex.double([11.0, 12.0, 13.0, 14.0, 15.0, 16.0])
    r["intensity.scale.value"] = flex.double([21.0, 22.0, 23.0, 24.0, 25.0, 26.0])
    r["intensity.prf.variance"] = flex.double([0.1, 0.2, 0.3, 0.4, 0.5, 0.6])
    r["intensity.sum.variance"] = flex.double([1.1, 1.2, 1.3, 1.4, 1.5, 1.6])
    r["intensity.scale.variance"] = flex.double([2.1, 2.2, 2.3, 2.4, 2.5, 2.6])
    r["inverse_scale_factor"] = flex.double([5.0, 5.0, 5.0, 10.0, 10.0, -10.0])
    r["inverse_scale_factor_variance"] = flex.double([1.0, 1.0, 1.0, 2.0, 2.0, 2.0])
    r.set_flags(
        flex.bool([True, True, True, False, False, True]), r.flags.integrated_prf
    )
    r.set_flags(
        flex.bool([True, True, True, True, False, False]), r.flags.integrated_sum
    )
    r.set_flags(
        flex.bool([True, False, False, False, False, False]),
        r.flags.excluded_for_scaling,
    )
    r.set_flags(
        flex.bool([False, True, False, False, False, False]), r.flags.outlier_in_scaling
    )
    bad = r.get_flags(r.flags.bad_for_scaling, all=False)
    r.set_flags(~bad, r.flags.scaled)
    r["id"] = flex.int(6, 0)
    r["partial_id"] = flex.int([0, 1, 2, 3, 4, 5])
    return r


def generate_test_reflections_for_scaling():
    """Generate a test table for scaling corrections."""
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0])
    r["intensity.sum.value"] = flex.double([11.0, 12.0, 13.0, 14.0, 15.0, 16.0, 17.0])
    r["intensity.scale.value"] = flex.double([21.0, 22.0, 23.0, 24.0, 25.0, 26.0, 27.0])
    r["intensity.prf.variance"] = flex.double([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
    r["intensity.sum.variance"] = flex.double([1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7])
    r["intensity.scale.variance"] = flex.double([2.1, 2.2, 2.3, 2.4, 2.5, 2.6, 2.7])
    r["inverse_scale_factor"] = flex.double([5.0, 5.0, 5.0, 10.0, 10.0, 10.0, 10.0])
    r["inverse_scale_factor_variance"] = flex.double(
        [1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0]
    )
    r["lp"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0, 0.5, 1.0])
    r["qe"] = flex.double([1.0, 1.0, 1.0, 1.0, 0.25, 1.0, 0.0])
    r["partiality"] = flex.double([0.1, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    r["id"] = flex.int(7, 0)
    r["partial_id"] = flex.int([0, 1, 2, 3, 4, 5, 6])
    return r


fpath = "dials.util.filter_reflections"


def test_filtered_arrays_from_experiments_reflections():
    """Test the creating of a miller array from crystal and reflection table."""
    refl = generate_integrated_test_reflections()
    refl["miller_index"] = flex.miller_index(
        [(1, 0, 0), (2, 0, 0), (3, 0, 0), (4, 0, 0), (5, 0, 0), (6, 0, 0)]
    )
    experiments = ExperimentList()
    exp_dict = {
        "__id__": "crystal",
        "real_space_a": [1.0, 0.0, 0.0],
        "real_space_b": [0.0, 1.0, 0.0],
        "real_space_c": [0.0, 0.0, 2.0],
        "space_group_hall_symbol": " C 2y",
    }
    crystal = Crystal.from_dict(exp_dict)
    experiments.append(Experiment(crystal=crystal))

    miller_set = filtered_arrays_from_experiments_reflections(experiments, [refl])[0]
    assert isinstance(miller_set, miller.set)
    assert list(miller_set.data()) == [4.6, 2.4, 2.5]  # same as calling filter
    # for export on scale intensity reducer.
    # now try for prf
    del refl["intensity.scale.value"]
    miller_set = filtered_arrays_from_experiments_reflections(experiments, [refl])[0]
    assert isinstance(miller_set, miller.set)
    assert list(miller_set.data()) == [1.0, 2.0, 3.0]  # same as calling filter
    # for export on prf + sum intensity reducer.
    # now just for sum
    del refl["intensity.prf.value"]
    miller_set = filtered_arrays_from_experiments_reflections(experiments, [refl])[0]
    assert isinstance(miller_set, miller.set)
    assert list(miller_set.data()) == [11.0, 12.0, 13.0, 14.0]  # same as calling
    # filter for export on prf intensity reducer.

    # Now try with a bad dataset - should be filtered.
    refl = generate_integrated_test_reflections()
    refl["miller_index"] = flex.miller_index(
        [(1, 0, 0), (2, 0, 0), (3, 0, 0), (4, 0, 0), (5, 0, 0), (6, 0, 0)]
    )
    # Trigger filtering on prf/sum, but when prf is bad - should proceed with sum
    refl.unset_flags(flex.bool(6, True), refl.flags.integrated_prf)
    del refl["intensity.scale.value"]
    refl2 = generate_integrated_test_reflections()
    refl2["partiality"] = flex.double(6, 0.0)
    experiments = ExperimentList()
    experiments.append(Experiment(crystal=crystal))
    experiments.append(Experiment(crystal=crystal))
    miller_sets = filtered_arrays_from_experiments_reflections(
        experiments, [refl, refl2], outlier_rejection_after_filter=True
    )
    assert len(miller_sets) == 1

    with pytest.raises(AssertionError):
        miller_sets = filtered_arrays_from_experiments_reflections(
            experiments, [refl, refl2], return_batches=True
        )

    experiments = ExperimentList()
    experiments.append(Experiment(crystal=crystal))
    experiments.append(Experiment(crystal=crystal))
    refl2 = generate_integrated_test_reflections()
    refl2["partiality"] = flex.double(6, 0.0)
    with pytest.raises(ValueError):
        refl["partiality"] = flex.double(6, 0.0)
        _ = filtered_arrays_from_experiments_reflections(experiments, [refl, refl2])


def test_filtered_arrays_from_experiments_reflections_with_batches(dials_data):
    x4wide_dir = dials_data("x4wide_processed", pathlib=True)
    refl = flex.reflection_table.from_file(x4wide_dir / "AUTOMATIC_DEFAULT_scaled.refl")
    expts = ExperimentListFactory.from_serialized_format(
        x4wide_dir / "AUTOMATIC_DEFAULT_scaled.expt", check_format=False
    )
    bad = refl.get_flags(refl.flags.bad_for_scaling, all=False)
    refl.set_flags(~bad, refl.flags.scaled)
    miller_arrays, batches = filtered_arrays_from_experiments_reflections(
        expts, [refl], return_batches=True
    )
    assert len(miller_arrays) == len(batches) == 1
    assert miller_arrays[0].size() == batches[0].size() == 62920
    assert len(set(batches[0].data())) == 88
    assert flex.min(batches[0].data()) == 2
    assert flex.max(batches[0].data()) == 89


def test_IntensityReducer_instantiations():
    """Test that all classes can be instantiated (have the required implemented
    methods) and have an intensities list with at least one str values"""
    allowed_intensities = FilterForExportAlgorithm.allowed_intensities
    SumIntensityReducer()
    assert all(i in allowed_intensities for i in SumIntensityReducer.intensities)
    PrfIntensityReducer()
    assert all(i in allowed_intensities for i in PrfIntensityReducer.intensities)
    SumAndPrfIntensityReducer()
    assert all(i in allowed_intensities for i in SumAndPrfIntensityReducer.intensities)
    ScaleIntensityReducer()
    assert all(i in allowed_intensities for i in ScaleIntensityReducer.intensities)


def test_FilteringReductionMethods():
    """Test the FilteringReductionMethods class."""
    # Test ice ring filtering
    reflections = generate_simple_table()
    reflections = FilteringReductionMethods.filter_ice_rings(reflections)
    assert list(reflections["intensity.prf.value"]) == [2.0, 3.0]

    # Test filtering on I/sigI
    reflections = generate_simple_table()
    reflections = FilteringReductionMethods._filter_on_min_isigi(
        reflections, "prf", 2.5
    )
    assert list(reflections["intensity.prf.value"]) == [3.0]

    # Test bad variance filtering
    reflections = generate_simple_table()
    reflections["intensity.prf.variance"][0] = 0.0
    reflections = FilteringReductionMethods._filter_bad_variances(reflections, "prf")
    assert list(reflections["intensity.prf.value"]) == [2.0, 3.0]

    # Test filtering on dmin
    reflections = generate_simple_table()
    reflections["d"] = flex.double([1.0, 2.0, 3.0])
    reflections = FilteringReductionMethods.filter_on_d_min(reflections, 1.5)
    assert list(reflections["d"]) == [2.0, 3.0]

    # test calculate_lp_qe_correction_and_filter - should be lp/qe
    # cases, qe, dqe , lp , qe zero
    r = flex.reflection_table()
    r["data"] = flex.double([1.0, 2.0, 3.0])
    r, c = FilteringReductionMethods.calculate_lp_qe_correction_and_filter(r)
    assert list(c) == pytest.approx([1.0, 1.0, 1.0])

    r["lp"] = flex.double([1.0, 0.5, 1.0])
    r, c = FilteringReductionMethods.calculate_lp_qe_correction_and_filter(r)
    assert list(c) == pytest.approx([1.0, 0.5, 1.0])

    r["qe"] = flex.double([0.25, 1.0, 0.0])
    r, c = FilteringReductionMethods.calculate_lp_qe_correction_and_filter(r)
    assert list(c) == pytest.approx([4.0, 0.5])
    del r["qe"]
    r["dqe"] = flex.double([0.25, 0.0])
    r, c = FilteringReductionMethods.calculate_lp_qe_correction_and_filter(r)
    assert list(c) == pytest.approx([4.0])

    # test filter unassigned
    r = flex.reflection_table()
    r["id"] = flex.int([-1, 0])
    r["i"] = flex.double([1.0, 2.0])
    r = FilteringReductionMethods.filter_unassigned_reflections(r)
    assert list(r["i"]) == [2.0]

    with mock.patch(
        fpath + ".sum_partial_reflections", side_effect=return_reflections_side_effect
    ) as sum_partials:
        reflections = generate_simple_table()
        reflections = FilteringReductionMethods.combine_and_filter_partials(
            reflections, partiality_threshold=0.7
        )
        assert sum_partials.call_count == 1
        assert list(reflections["intensity.prf.value"]) == [1.0, 2.0]
        reflections = generate_simple_table()
        FilteringReductionMethods.combine_and_filter_partials(
            reflections, partiality_threshold=0.4
        )
        assert sum_partials.call_count == 2
        assert list(reflections["intensity.prf.value"]) == [1.0, 2.0, 3.0]


def test_PrfIntensityReducer():
    """Test the methods of the PrfIntensityReducer class"""
    reflections = generate_integrated_test_reflections()
    reflections = PrfIntensityReducer.reduce_on_intensities(reflections)
    assert list(reflections["intensity.prf.value"]) == [1.0, 2.0, 3.0, 6.0]
    assert list(reflections["intensity.prf.variance"]) == [0.1, 0.2, 0.3, 0.6]

    reflections = generate_test_reflections_for_scaling()
    reflections = PrfIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.prf.value"]) == [1.0, 3.0, 4.0, 20.0, 3.0]
    assert list(reflections["intensity.prf.variance"]) == pytest.approx(
        [0.1, 0.3, 0.4, 0.5 * 16.0, 0.6 * 0.25]
    )

    # check it still passes with no partiality correction
    reflections = generate_test_reflections_for_scaling()
    del reflections["partiality"]
    reflections = PrfIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.prf.value"]) == [1.0, 2.0, 3.0, 4.0, 20.0, 3.0]

    # test IsgiI selecting
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0])
    r["intensity.prf.variance"] = flex.double([1.0, 1.0])
    r = PrfIntensityReducer.filter_on_min_isigi(r, 1.5)
    assert list(r["intensity.prf.value"]) == [2.0]

    # now test a typical case - reflection table with inv scale factor,
    # check that we only apply that correction and do any relevant filtering
    reflections = generate_integrated_test_reflections()
    reflections["qe"] = flex.double(6, 2)
    reflections = PrfIntensityReducer.filter_for_export(reflections)

    assert list(reflections["intensity.prf.value"]) == pytest.approx(
        [1.0 / 2.0, 2.0 / 2.0, 3.0 / 2.0, 6.0 / 2.0]
    )
    assert list(reflections["intensity.prf.variance"]) == pytest.approx(
        [0.1 / 4.0, 0.2 / 4.0, 0.3 / 4.0, 0.6 / 4.0]
    )

    assert list(reflections["fractioncalc"]) == [1.0] * 4

    assert "intensity.sum.value" not in reflections
    assert "intensity.scale.value" not in reflections
    assert "intensity.sum.variance" not in reflections
    assert "intensity.scale.variance" not in reflections


def test_SumIntensityReducer():
    """Test the methods of the SumIntensityReducer class"""
    reflections = generate_integrated_test_reflections()
    reflections = SumIntensityReducer.reduce_on_intensities(reflections)
    # reflections = reduce_on_summation_intensities(reflections)
    assert list(reflections["intensity.sum.value"]) == [11.0, 12.0, 13.0, 14.0]
    assert list(reflections["intensity.sum.variance"]) == [1.1, 1.2, 1.3, 1.4]

    reflections = generate_test_reflections_for_scaling()
    reflections = SumIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.sum.value"]) == [110.0, 13.0, 14.0, 60.0, 8.0]
    assert list(reflections["intensity.sum.variance"]) == pytest.approx(
        [110.0, 1.3, 1.4, 24.0, 0.4]
    )

    # check it still passes with no partiality correction
    reflections = generate_test_reflections_for_scaling()
    del reflections["partiality"]
    reflections = SumIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.sum.value"]) == [
        11.0,
        12.0,
        13.0,
        14.0,
        60.0,
        8.0,
    ]

    # test IsgiI selecting
    r = flex.reflection_table()
    r["intensity.sum.value"] = flex.double([1.0, 2.0])
    r["intensity.sum.variance"] = flex.double([1.0, 1.0])
    r = SumIntensityReducer.filter_on_min_isigi(r, 1.5)
    assert list(r["intensity.sum.value"]) == [2.0]

    # now test a typical case - reflection table with inv scale factor,
    # check that we only apply that correction and do any relevant filtering
    reflections = generate_integrated_test_reflections()
    reflections["lp"] = flex.double(6, 2)
    reflections["partiality"] = flex.double([1.0, 1.0, 0.5, 0.1, 1.0, 1.0])
    reflections["d"] = flex.double([5.0, 5.0, 0.1, 5.0, 5.0, 5.0])
    reflections = SumIntensityReducer.filter_for_export(
        reflections, partiality_threshold=0.25, d_min=1.0
    )

    assert list(reflections["intensity.sum.value"]) == pytest.approx([22.0, 24.0])
    assert list(reflections["intensity.sum.variance"]) == pytest.approx([4.4, 4.8])

    assert "intensity.prf.value" not in reflections
    assert "intensity.scale.value" not in reflections
    assert "intensity.prf.variance" not in reflections
    assert "intensity.scale.variance" not in reflections


def test_SumAndPrfIntensityReducer():
    """Test that the reflection table is reduced on prf and sum intensities"""
    reflections = generate_integrated_test_reflections()
    reflections = SumAndPrfIntensityReducer.reduce_on_intensities(reflections)
    assert list(reflections["intensity.prf.value"]) == [1.0, 2.0, 3.0]
    assert list(reflections["intensity.prf.variance"]) == [0.1, 0.2, 0.3]
    assert list(reflections["intensity.sum.value"]) == [11.0, 12.0, 13.0]
    assert list(reflections["intensity.sum.variance"]) == [1.1, 1.2, 1.3]

    reflections["lp"] = flex.double([0.5, 0.5, 1.0])
    reflections["partiality"] = flex.double([0.0, 1.0, 1.0])
    reflections = SumAndPrfIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.prf.value"]) == [1.0, 3.0]
    assert list(reflections["intensity.prf.variance"]) == [0.2 * 0.25, 0.3]
    assert list(reflections["intensity.sum.value"]) == [6.0, 13.0]
    assert list(reflections["intensity.sum.variance"]) == [0.3, 1.3]

    # test IsgiI selecting
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.prf.variance"] = flex.double([1.0, 1.0, 1.0])
    r["intensity.sum.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.sum.variance"] = flex.double([1.0, 1.1, 1.0])
    r = SumAndPrfIntensityReducer.filter_on_min_isigi(r, 1.999)
    assert list(r["intensity.prf.value"]) == [3.0]

    # test bad variance filtering
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.prf.variance"] = flex.double([0.0, 1.0, 1.0])
    r["intensity.sum.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.sum.variance"] = flex.double([1.0, 0.0, 1.0])
    r.set_flags(flex.bool([True, True, True]), r.flags.integrated_prf)
    r.set_flags(flex.bool([True, True, True]), r.flags.integrated_sum)
    r = SumAndPrfIntensityReducer.filter_bad_variances(r)
    assert list(r["intensity.prf.value"]) == [3.0]

    # test filtering for export
    reflections = generate_integrated_test_reflections()
    reflections = SumAndPrfIntensityReducer.filter_for_export(reflections)
    assert "intensity.scale.value" not in reflections
    assert "intensity.scale.variance" not in reflections


def test_ScaleIntensityReducer():
    """Test that the reflection table is reduced on scaling intensities"""
    reflections = generate_integrated_test_reflections()
    reflections = ScaleIntensityReducer.reduce_on_intensities(reflections)
    assert list(reflections["intensity.scale.value"]) == pytest.approx(
        [23.0, 24.0, 25.0]
    )
    assert list(reflections["intensity.scale.variance"]) == pytest.approx(
        [2.3, 2.4, 2.5]
    )
    del reflections["inverse_scale_factor"]
    with pytest.raises(AssertionError):
        reflections = ScaleIntensityReducer.reduce_on_intensities(reflections)

    reflections = generate_test_reflections_for_scaling()
    reflections = ScaleIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.scale.value"]) == pytest.approx(
        [21.0 / 5.0, 23.0 / 5.0, 24.0 / 10.0, 25.0 / 10.0, 26.0 / 10.0, 27.0 / 10.0]
    )
    assert list(reflections["intensity.scale.variance"]) == pytest.approx(
        [2.1 / 25.0, 2.3 / 25.0, 2.4 / 100, 2.5 / 100, 2.6 / 100, 2.7 / 100]
    )
    del reflections["inverse_scale_factor"]
    with pytest.raises(AssertionError):
        reflections = ScaleIntensityReducer.apply_scaling_factors(reflections)

    reflections = generate_test_reflections_for_scaling()
    del reflections["partiality"]
    reflections = ScaleIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.scale.value"]) == [
        21.0 / 5.0,
        22.0 / 5.0,
        23.0 / 5.0,
        24.0 / 10.0,
        25.0 / 10.0,
        26.0 / 10.0,
        27.0 / 10.0,
    ]

    # test IsgiI selecting
    r = flex.reflection_table()
    r["intensity.scale.value"] = flex.double([1.0, 2.0])
    r["intensity.scale.variance"] = flex.double([1.0, 1.0])
    r = ScaleIntensityReducer.filter_on_min_isigi(r, 1.5)
    assert list(r["intensity.scale.value"]) == [2.0]

    # now test a typical case - reflection table with inv scale factor,
    # check that we only apply that correction and do any relevant filtering
    reflections = generate_integrated_test_reflections()
    reflections["lp"] = flex.double(6, 0.6)
    reflections = ScaleIntensityReducer.filter_for_export(reflections)

    assert list(reflections["intensity.scale.value"]) == pytest.approx(
        [23.0 / 5.0, 24.0 / 10.0, 25.0 / 10.0]
    )
    assert list(reflections["intensity.scale.variance"]) == pytest.approx(
        [2.3 / 25.0, 0.024, 0.025]
    )

    assert "intensity.prf.value" not in reflections
    assert "intensity.sum.value" not in reflections
    assert "intensity.prf.variance" not in reflections
    assert "intensity.sum.variance" not in reflections


def test_AllSumPrfScaleIntensityReducer():
    reflections = generate_integrated_test_reflections()
    reflections = AllSumPrfScaleIntensityReducer.reduce_on_intensities(reflections)
    assert list(reflections["intensity.scale.value"]) == [23.0]
    assert list(reflections["intensity.scale.variance"]) == [2.3]
    assert list(reflections["intensity.prf.value"]) == [3.0]
    assert list(reflections["intensity.prf.variance"]) == [0.3]
    assert list(reflections["intensity.sum.value"]) == [13.0]
    assert list(reflections["intensity.sum.variance"]) == [1.3]
    reflections = AllSumPrfScaleIntensityReducer.apply_scaling_factors(reflections)
    assert list(reflections["intensity.scale.value"]) == [23.0 / 5.0]
    assert list(reflections["intensity.scale.variance"]) == [2.3 / 25.0]
    assert list(reflections["intensity.prf.value"]) == [3.0]
    assert list(reflections["intensity.prf.variance"]) == [0.3]
    assert list(reflections["intensity.sum.value"]) == [13.0]
    assert list(reflections["intensity.sum.variance"]) == [1.3]


def test_filter_reflection_table():
    """Test the interface function"""
    reflections = generate_integrated_test_reflections()
    reflections = filter_reflection_table(reflections, ["sum"])
    assert "intensity.sum.value" in reflections
    assert "intensity.prf.value" not in reflections
    assert "intensity.scale.value" not in reflections
    reflections = generate_integrated_test_reflections()
    reflections = filter_reflection_table(reflections, ["profile"])
    assert "intensity.prf.value" in reflections
    assert "intensity.sum.value" not in reflections
    assert "intensity.scale.value" not in reflections
    reflections = generate_integrated_test_reflections()
    reflections = filter_reflection_table(reflections, ["scale"])
    assert "intensity.scale.value" in reflections
    assert "intensity.prf.value" not in reflections
    assert "intensity.sum.value" not in reflections
    reflections = generate_integrated_test_reflections()
    reflections = filter_reflection_table(reflections, ["sum", "profile"])
    assert "intensity.sum.value" in reflections
    assert "intensity.prf.value" in reflections
    assert "intensity.scale.value" not in reflections
    reflections = generate_integrated_test_reflections()
    reflections = filter_reflection_table(reflections, ["sum", "profile", "scale"])
    assert "intensity.sum.value" in reflections
    assert "intensity.prf.value" in reflections
    assert "intensity.scale.value" in reflections
    with pytest.raises(ValueError):
        reflections = filter_reflection_table(reflections, ["bad"])

    # try filter with profile option and no profiles
    reflections = generate_integrated_test_reflections()
    reflections.unset_flags(
        flex.bool(reflections.size(), True), reflections.flags.integrated_prf
    )
    reflections = filter_reflection_table(reflections, ["sum", "profile"])
    # should try profile but fail and so retry with just sum
    assert "intensity.sum.value" in reflections
    assert "intensity.prf.value" not in reflections

    reflections = generate_integrated_test_reflections()
    reflections.unset_flags(
        flex.bool(reflections.size(), True), reflections.flags.integrated_prf
    )
    reflections = filter_reflection_table(reflections, ["sum", "profile", "scale"])
    # should try profile but fail and so retry with just sum
    assert "intensity.sum.value" in reflections
    assert "intensity.prf.value" not in reflections
    assert "intensity.scale.value" in reflections

    reflections = generate_integrated_test_reflections()
    reflections.unset_flags(
        flex.bool(reflections.size(), True), reflections.flags.integrated_prf
    )
    with pytest.raises(ValueError):
        _ = filter_reflection_table(reflections, ["profile"])


def return_reflections_side_effect(reflections, *args, **kwargs):
    """Side effect for overriding the call to reject_outliers."""
    return reflections


def test_checks_in_reduce_data_for_export():
    """Test for graceful failures if serious problems with the data are found."""
    # If no valid ids, should raise sorry
    r = flex.reflection_table()
    r["id"] = flex.int([-1, -1, -1])
    r["intensity.prf.value"] = flex.double(3, 1.0)
    r["intensity.prf.variance"] = flex.double(3, 1.0)
    with pytest.raises(ValueError):
        r = PrfIntensityReducer.filter_for_export(r)

    # If no valid prf, should raise NoProfilesException
    r = flex.reflection_table()
    r["id"] = flex.int([0, 0])
    r["intensity.prf.value"] = flex.double([1.0, 2.0])
    r["intensity.prf.variance"] = flex.double([0.0, 1.0])
    r.set_flags(flex.bool([False, False]), r.flags.integrated_prf)
    with pytest.raises(NoProfilesException):
        r = PrfIntensityReducer.filter_for_export(r)
    r.set_flags(flex.bool([True, True]), r.flags.integrated_prf)

    # Should filter bad variances
    r = PrfIntensityReducer.filter_bad_variances(r)
    assert r.size() == 1

    # If reflection table is empty
    r = flex.reflection_table()
    with pytest.raises(AssertionError):
        r = PrfIntensityReducer.filter_for_export(r)

    # What if all ice ring
    r = generate_simple_table()
    r.set_flags(flex.bool([True, True, True]), r.flags.in_powder_ring)
    with pytest.raises(ValueError):
        _ = PrfIntensityReducer.filter_for_export(
            r, filter_ice_rings=True, min_isigi=1.0, partiality_threshold=0.99
        )

    # What if none left on SigI
    r = generate_simple_table()
    with pytest.raises(ValueError):
        _ = PrfIntensityReducer.filter_for_export(
            r, filter_ice_rings=False, min_isigi=4.0, partiality_threshold=0.99
        )


def test_partial_summing_functions():
    """These functions set the first entry to be the sum/weighted sum for the
    partials. Example is that of two reflections measured with 0.5 partiality,
    with I/sigma of 1 & 2."""
    # FIXME: What weights should be used
    # Test combining scale values
    r = flex.reflection_table()
    r["intensity.scale.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    r["intensity.scale.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0])
    partials_list = [0, 1]
    r = _sum_scale_partials(r, partials_list)
    assert r.size() == 5
    assert r["intensity.scale.value"][0] == 1.5
    assert r["intensity.scale.variance"][0] == 0.5

    # Test combining prf values
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    r["intensity.prf.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0])
    partials_list = [0, 1]
    r = _sum_prf_partials(r, partials_list)
    assert r.size() == 5
    assert r["intensity.prf.value"][0] == 1.8
    assert r["intensity.prf.variance"][0] == 1.0

    # Test summing sum values
    # "Same data - sum values should be half, adjust variances accordingly
    # to make I/sigma = 1,2"
    r = flex.reflection_table()
    r["intensity.sum.value"] = flex.double([0.5, 1.0, 3.0, 4.0, 5.0])
    r["intensity.sum.variance"] = flex.double([0.25, 0.25, 1.0, 1.0, 1.0])
    partials_list = [0, 1]
    r = _sum_sum_partials(r, partials_list)
    assert r.size() == 5
    assert r["intensity.sum.value"][0] == 1.5
    assert r["intensity.sum.variance"][0] == 0.5


def test_sum_partial_reflections():
    """Test the partial summing function for the different intensity types."""
    r = flex.reflection_table()
    r["intensity.prf.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    r["intensity.prf.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0])
    r["partial_id"] = flex.int([0, 0, 1, 1, 2])
    r["partiality"] = flex.double([0.5, 0.4, 0.6, 0.2, 0.9])
    r["identifier"] = flex.int([1, 2, 3, 4, 5])

    r = sum_partial_reflections(r)
    assert list(r["identifier"]) == [1, 3, 5]
    assert list(r["partiality"]) == [0.9, 0.8, 0.9]

    r = flex.reflection_table()
    r["intensity.sum.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    r["intensity.sum.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0])
    r["partial_id"] = flex.int([0, 0, 1, 1, 2])
    r["partiality"] = flex.double([0.5, 0.4, 0.6, 0.2, 0.9])
    r["identifier"] = flex.int([1, 2, 3, 4, 5])

    r = sum_partial_reflections(r)
    assert list(r["identifier"]) == [1, 3, 5]
    assert list(r["partiality"]) == [0.9, 0.8, 0.9]

    r = flex.reflection_table()
    r["intensity.scale.value"] = flex.double([1.0, 2.0, 3.0, 4.0, 5.0])
    r["intensity.scale.variance"] = flex.double([1.0, 1.0, 1.0, 1.0, 1.0])
    r["partial_id"] = flex.int([0, 0, 1, 1, 2])
    r["partiality"] = flex.double([0.5, 0.4, 0.6, 0.2, 0.9])
    r["identifier"] = flex.int([1, 2, 3, 4, 5])

    r = sum_partial_reflections(r)
    assert list(r["identifier"]) == [1, 3, 5]
    assert list(r["partiality"]) == [0.9, 0.8, 0.9]

    # if all partiality of one - should just return same
    r = flex.reflection_table()
    r["intensity.scale.value"] = flex.double([1.0, 2.0, 3.0])
    r["intensity.scale.variance"] = flex.double([1.0, 1.0, 1.0])
    r["partiality"] = flex.double(3, 1.0)
    r["identifier"] = flex.int([1, 2, 3])
    r2 = sum_partial_reflections(r)
    assert r2 is r
    assert list(r["identifier"]) == [1, 2, 3]

    # Add test to check calculation in case where both prf and sum - but this
    # requires knowing how the values will be weighted, so leave until that is
    # decided.



from __future__ import annotations

import pickle


class ModelCache:
    """
    A class to cache the model
    """

    def __init__(self):
        """
        Create a model dictionary
        """
        self.model = {}

    def get(self, name):
        """
        Get the model
        """
        if name is None:
            raise RuntimeError("Model is not specified")
        try:
            model = self.model[name]
        except KeyError:
            with open(name, "rb") as infile:
                model = pickle.load(infile)
                self.model[name] = model
        return model


# Instance of the model cache
global_model_cache = ModelCache()


class BackgroundAlgorithm:
    """Class to do background subtraction."""

    def __init__(
        self,
        experiments,
        model=None,
        robust=False,
        tuning_constant=1.345,
        min_pixels=10,
    ):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param model: The background model
        :param robust: Use the robust background algorithm
        :param tuning_constant: The robust tuning constant
        """
        from dials.algorithms.background.gmodel import Creator

        # Get the model
        model = global_model_cache.get(model)

        # Create the background creator
        self._create = Creator(model=model, robust=robust, min_pixels=min_pixels)

    def compute_background(self, reflections, image_volume=None):
        """
        Compute the background.

        :param reflections: The list of reflections
        """
        # Do the background subtraction
        if image_volume is None:
            success = self._create(reflections)
            reflections["background.mean"] = reflections[
                "shoebox"
            ].mean_modelled_background()
        else:
            success = self._create(reflections, image_volume)
        reflections.set_flags(~success, reflections.flags.dont_integrate)
        return success


class GModelBackgroundCalculatorFactory:
    """Class to do background subtraction."""

    @staticmethod
    def create(experiments, model=None, robust=False, min_pixels=10):
        """
        Initialise the algorithm.

        :param experiments: The list of experiments
        :param model: The background model
        :param robust: Use the robust background algorithm
        :param tuning_constant: The robust tuning constant
        """
        from dials.algorithms.integration.parallel_integrator import (
            GModelBackgroundCalculator,
        )

        # Get the model
        model = global_model_cache.get(model)

        # Create the background creator
        return GModelBackgroundCalculator(
            model=model, robust=robust, min_pixels=min_pixels
        )



from __future__ import annotations

import functools
import json
import operator
import os
import pathlib
import stat
import sys
from distutils.version import LooseVersion as parse_version

import pkg_resources

import libtbx.load_env

try:
    import conda.cli.python_api
except ModuleNotFoundError:
    conda = None

try:
    import pre_commit.constants
except ModuleNotFoundError:
    pre_commit = None

BOLD = "\033[1m"
GREEN = "\033[32m"
MAGENTA = "\033[1;35m"
NC = "\033[0m"
RED = "\033[1;31m"
YELLOW = "\033[1;33m"

minimum_conda_version = "4.8.0"
minimum_precommit_version = "2.16"
minimum_python_version = "3.8.0"

repo_prefix = "  {:.<15}:"
repo_no_precommit = "(no pre-commit hooks)"
repo_precommit_installed = GREEN + "pre-commit installed" + NC
repo_precommit_conflict = (
    RED + "pre-commit available but a different pre-commit hook is installed" + NC
)
repo_precommit_legacy = YELLOW + "pre-commit out of date" + NC
repo_precommit_available = MAGENTA + "pre-commit available but not installed" + NC
repo_is_worktree = "(is a git worktree checkout)"


def precommitbx_template():
    prefix_path = pathlib.Path(_conda_info()["default_prefix"])
    assert prefix_path.is_dir(), f"Detected conda prefix path {prefix_path} is invalid"
    activate_script = prefix_path / "etc" / "profile.d" / "conda.sh"
    assert (
        activate_script.is_file()
    ), f"Expected conda activation script {activate_script} not found"
    file_content = (
        "#!/bin/bash",
        "# File generated by precommitbx for conda environments",
        "export LD_LIBRARY_PATH=",
        "export PYTHONPATH=",
        "export SETUPTOOLS_USE_DISTUTILS=stdlib",
        f"export PRE_COMMIT_HOME={prefix_path}/.precommit",
        "if [ ! -f .pre-commit-config.yaml ]; then",
        "  echo No pre-commit configuration. Skipping pre-commit checks.",
        "  exit 0",
        "fi",
        f"source {activate_script}",
        f"conda activate {prefix_path}",
        "pre-commit run --config .pre-commit-config.yaml --hook-stage commit",
    )
    return "\n".join(file_content) + "\n"


def install_precommitbx_hook(path):
    with path.joinpath(".git", "hooks", "pre-commit").open("w") as fh:
        fh.write(precommitbx_template())
        if os.name != "nt":
            mode = os.fstat(fh.fileno()).st_mode
            mode |= stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
            os.fchmod(fh.fileno(), stat.S_IMODE(mode))


def check_precommitbx_hook(path):
    hookfile = path / ".git" / "hooks" / "pre-commit"
    if not hookfile.is_file():
        return False
    if not os.access(hookfile, os.X_OK):
        return False
    hook = hookfile.read_text()
    if hook == precommitbx_template():
        return repo_precommit_installed
    if "generated by precommitbx" in hook:
        return repo_precommit_legacy
    if "DIALS_WITHOUT_PRECOMMITS" in hook:
        return False
    if hook:
        return repo_precommit_conflict
    return False


@functools.lru_cache()
def _conda_info():
    conda_info, error, return_code = conda.cli.python_api.run_command(
        conda.cli.python_api.Commands.INFO,
        "--json",
        use_exception_handler=True,
    )
    if return_code:
        if conda_info and conda_info.get("message"):
            print(conda_info["message"])
        raise RuntimeError(error)
    return json.loads(conda_info)


def _install_precommit():
    conda_info, error, return_code = conda.cli.python_api.run_command(
        conda.cli.python_api.Commands.INSTALL,
        "--json",
        "--quiet",
        "--yes",
        "pre-commit",
        use_exception_handler=True,
    )
    if return_code:
        raise RuntimeError(error)
    information = json.loads(conda_info)
    if not information.get("success"):
        print(information)
        raise RuntimeError("Could not install precommit into conda environment")
    installed_packages = information.get("actions", {}).get("LINK", [])
    precommit_version = None
    for package in sorted(installed_packages, key=operator.itemgetter("name")):
        print(f"  installed {package['name']} {package['version']}")
        if package["name"] == "pre-commit":
            precommit_version = package["version"]
    return precommit_version


def list_all_repository_candidates():
    repositories = {}
    for module in sorted(libtbx.env.module_dict):
        module_paths = [
            pathlib.Path(abs(path))
            for path in libtbx.env.module_dict[module].dist_paths
            if path and (path / ".git").exists()
        ]
        if not module_paths:
            continue
        if len(module_paths) == 1:
            repositories[module] = module_paths[0]
        else:
            for path in module_paths:
                repositories[f"{module}:{path}"] = path
    for ep in pkg_resources.iter_entry_points("libtbx.precommit"):
        path = pathlib.Path(ep.load().__path__[0])
        if path.joinpath(".git").is_dir():
            repositories[ep.name] = path
        elif path.parent.joinpath(".git").is_dir():
            repositories[ep.name] = path.parent
    return repositories


def _version_check(package, version, minimum):
    if parse_version(version) < parse_version(minimum):
        colour = YELLOW
        expected = f" (expected: {minimum} or higher)"
    else:
        colour = GREEN
        expected = ""
    print(f"{package:10s}: {colour}{version}{NC}{expected}")
    return colour == GREEN


def main():
    install_things = "install" in sys.argv

    install_precommit = False
    if _conda_info():
        _version_check("Conda", _conda_info()["conda_version"], minimum_conda_version)
    else:
        print(f"Conda:      {RED}could not read information{NC}")
    if pre_commit:
        if not _version_check(
            "Pre-commit", pre_commit.constants.VERSION, minimum_precommit_version
        ):
            install_precommit = True
    else:
        print(f"Pre-commit: {RED}not installed{NC}")
        install_precommit = True
    _version_check(
        "Python",
        f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}",
        minimum_python_version,
    )

    # Need to obtain list of repository candidates before potentially installing precommit
    # as this may jumble up installed packages while the python process is running.
    repositories = list_all_repository_candidates()

    if install_things and install_precommit:
        if not _conda_info():
            exit(f"{RED}Can not install precommit without conda present{NC}")
        print("\nInstalling precommit...")
        precommit_version = _install_precommit()
        if not precommit_version:
            exit(f"{RED}Could not install precommit{NC}")
        try:
            import pre_commit.constants as pcc
        except ModuleNotFoundError:
            exit(
                f"Installation of precommit {GREEN}succeeded{NC}, {RED}but could not confirm version number{NC}"
            )
        if not _version_check("Pre-commit", pcc.VERSION, minimum_precommit_version):
            exit(
                f"Installation of precommit {GREEN}succeeded{NC}, {RED}but version is still insufficient{NC}"
            )

    print()
    print("Repositories:")
    changes_required = False
    for path in sys.argv[1:]:
        if path == "install":
            continue
        path = pathlib.Path(".").joinpath(path).absolute()
        if path.name in repositories:
            base = os.fspath(path)
        else:
            base = path.name
        repositories[base] = path
    for module in sorted(repositories):
        if not repositories[module].joinpath(".pre-commit-config.yaml").is_file():
            print(repo_prefix.format(module), repo_no_precommit)
            continue
        if repositories[module].joinpath(".git").is_file():
            print(repo_prefix.format(module), repo_is_worktree)
            continue
        message = (
            check_precommitbx_hook(repositories[module]) or repo_precommit_available
        )
        if message != repo_precommit_installed and install_things:
            install_precommitbx_hook(repositories[module])
            message = (
                check_precommitbx_hook(repositories[module]) or repo_precommit_available
            )
        print(repo_prefix.format(module), message)
        if message != repo_precommit_installed:
            changes_required = True

    if changes_required:
        print()
        exit(f"To install pre-commit hooks run {BOLD}libtbx.precommit install{NC}")



from __future__ import annotations

import math

from scitbx import lbfgs
from scitbx.array_family import flex


def optimise_basis_vectors(reciprocal_lattice_points, vectors):
    optimised = flex.vec3_double()
    for vector in vectors:
        minimised = BasisVectorMinimiser(reciprocal_lattice_points, vector)
        optimised.append(tuple(minimised.x))
    functionals = flex.double(minimised.target.compute_functional(v) for v in vectors)
    perm = flex.sort_permutation(functionals)
    optimised = optimised.select(perm)
    return optimised


# Optimise the initial basis vectors as per equation 11.4.3.4 of
# Otwinowski et al, International Tables Vol. F, chapter 11.4 pp. 282-295
class BasisVectorTarget:
    def __init__(self, reciprocal_lattice_points):
        self.reciprocal_lattice_points = reciprocal_lattice_points
        self._xyz_parts = self.reciprocal_lattice_points.parts()

    def compute_functional(self, vector):
        two_pi_S_dot_v = 2 * math.pi * self.reciprocal_lattice_points.dot(vector)
        return -flex.sum(flex.cos(two_pi_S_dot_v))

    def compute_functional_and_gradients(self, vector):
        assert len(vector) == 3
        two_pi_S_dot_v = 2 * math.pi * self.reciprocal_lattice_points.dot(vector)
        f = -flex.sum(flex.cos(two_pi_S_dot_v))
        sin_part = flex.sin(two_pi_S_dot_v)
        g = flex.double(
            [flex.sum(2 * math.pi * self._xyz_parts[i] * sin_part) for i in range(3)]
        )
        return f, g


class BasisVectorMinimiser:
    def __init__(
        self,
        reciprocal_lattice_points,
        vector,
        lbfgs_termination_params=None,
        lbfgs_core_params=lbfgs.core_parameters(m=20),
    ):
        self.reciprocal_lattice_points = reciprocal_lattice_points
        if not isinstance(vector, flex.double):
            self.x = flex.double(vector)
        else:
            self.x = vector.deep_copy()
        self.n = len(self.x)
        assert self.n == 3
        self.target = BasisVectorTarget(self.reciprocal_lattice_points)
        self.minimizer = lbfgs.run(
            target_evaluator=self,
            termination_params=lbfgs_termination_params,
            core_params=lbfgs_core_params,
        )

    def compute_functional_and_gradients(self):
        f, g = self.target.compute_functional_and_gradients(tuple(self.x))
        return f, g



"""
Test combination of multiple experiments and reflections files.
"""


from __future__ import annotations

import copy
import os
from pathlib import Path

import procrunner
import pytest

from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.serialize import load

import dials.command_line.combine_experiments as combine_experiments
from dials.array_family import flex


def test(dials_regression, tmp_path):
    data_dir = Path(dials_regression) / "refinement_test_data" / "multi_narrow_wedges"

    input_range = list(range(2, 49))
    for i in (8, 10, 15, 16, 34, 39, 45):
        input_range.remove(i)

    phil_input = (
        "\n".join(
            (
                f"  input.experiments={data_dir}/data/sweep_{i:03d}/experiments.json\n"
                + f"  input.reflections={data_dir}/data/sweep_{i:03d}/reflections.pickle"
            )
            for i in input_range
        )
        + """
 reference_from_experiment.beam=0
 reference_from_experiment.scan=0
 reference_from_experiment.goniometer=0
 reference_from_experiment.detector=0
 """
    )
    tmp_path.joinpath("input.phil").write_text(phil_input)

    result = procrunner.run(
        ["dials.combine_experiments", "input.phil"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr

    # load results
    exp = ExperimentListFactory.from_json_file(
        tmp_path / "combined.expt", check_format=False
    )
    ref = flex.reflection_table.from_file(tmp_path / "combined.refl")

    # test the experiments
    assert len(exp) == 103
    assert len(exp.crystals()) == 103
    assert len(exp.beams()) == 1
    assert len(exp.scans()) == 1
    assert len(exp.detectors()) == 1
    assert len(exp.goniometers()) == 1
    for e in exp:
        assert e.imageset is not None

    # test the reflections
    assert len(ref) == 11689

    result = procrunner.run(
        ["dials.split_experiments", "combined.expt", "combined.refl"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    for i, e in enumerate(exp):
        assert tmp_path.joinpath(f"split_{i:03d}.expt").is_file()
        assert tmp_path.joinpath(f"split_{i:03d}.refl").is_file()

        exp_single = ExperimentListFactory.from_json_file(
            tmp_path / f"split_{i:03d}.expt", check_format=False
        )
        ref_single = flex.reflection_table.from_file(tmp_path / f"split_{i:03d}.refl")

        assert len(exp_single) == 1
        assert exp_single[0].crystal == e.crystal
        assert exp_single[0].beam == e.beam
        assert exp_single[0].detector == e.detector
        assert exp_single[0].scan == e.scan
        assert exp_single[0].goniometer == e.goniometer
        assert exp_single[0].imageset == e.imageset
        assert len(ref_single) == len(ref.select(ref["id"] == i))
        assert ref_single["id"].all_eq(0)

    result = procrunner.run(
        ["dials.split_experiments", "combined.expt", "output.experiments_prefix=test"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    for i in range(len(exp)):
        assert tmp_path.joinpath(f"test_{i:03d}.expt").is_file()

    # Modify a copy of the detector
    detector = copy.deepcopy(exp.detectors()[0])
    panel = detector[0]
    x, y, z = panel.get_origin()
    panel.set_frame(panel.get_fast_axis(), panel.get_slow_axis(), (x, y, z + 10))
    # Set half of the experiments to the new detector
    for i in range(len(exp) // 2):
        exp[i].detector = detector
    exp.as_json(tmp_path / "modded.expt")

    result = procrunner.run(
        [
            "dials.split_experiments",
            "modded.expt",
            "combined.refl",
            "output.experiments_prefix=test_by_detector",
            "output.reflections_prefix=test_by_detector",
            "by_detector=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    for i in range(2):
        assert tmp_path.joinpath(f"test_by_detector_{i:03d}.expt").is_file()
        assert tmp_path.joinpath(f"test_by_detector_{i:03d}.refl").is_file()
    assert not tmp_path.joinpath("test_by_detector_002.expt").is_file()
    assert not tmp_path.joinpath("test_by_detector_002.refl").is_file()

    # Now do test when input has identifiers set
    reflections = flex.reflection_table().from_file(tmp_path / "combined.refl")
    explist = ExperimentListFactory.from_json_file(
        tmp_path / "combined.expt", check_format=False
    )
    # set string identifiers as nonconsecutive 0,2,4,6....
    for i, exp in enumerate(explist):
        assert i in reflections["id"]
        reflections.experiment_identifiers()[i] = str(i * 2)
        exp.identifier = str(i * 2)
    reflections.as_file(tmp_path / "assigned.refl")
    explist.as_json(tmp_path / "assigned.expt")

    result = procrunner.run(
        ["dials.split_experiments", "assigned.expt", "assigned.refl"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    for i in range(len(explist)):
        assert tmp_path.joinpath(f"split_{i:03d}.expt").is_file()
        assert tmp_path.joinpath(f"split_{i:03d}.refl").is_file()

        exp_single = ExperimentListFactory.from_json_file(
            tmp_path / f"split_{i:03d}.expt", check_format=False
        )
        ref_single = flex.reflection_table.from_file(tmp_path / f"split_{i:03d}.refl")

        assert len(exp_single) == 1
        # resets all ids to 0, but keeps mapping to unique identifier.
        # doesn't have to be set to 0 but doing this to keep more consistent with
        # other dials programs
        assert ref_single["id"].all_eq(0)
        assert ref_single.experiment_identifiers()[0] == str(i * 2)

    # update modded experiments to have same identifiers as assigned_experiments
    moddedlist = ExperimentListFactory.from_json_file(
        tmp_path / "modded.expt", check_format=False
    )
    for i, exp in enumerate(moddedlist):
        exp.identifier = str(i * 2)
    moddedlist.as_json(tmp_path / "modded.expt")

    result = procrunner.run(
        [
            "dials.split_experiments",
            "modded.expt",
            "assigned.refl",
            "output.experiments_prefix=test_by_detector",
            "output.reflections_prefix=test_by_detector",
            "by_detector=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    # Expect each datasets to have ids from 0..50 with experiment identifiers
    # all kept from before 0,2,4,6,...
    current_exp_id = 0
    for i in range(2):
        assert tmp_path.joinpath(f"test_by_detector_{i:03d}.expt").is_file()
        assert tmp_path.joinpath(f"test_by_detector_{i:03d}.refl").is_file()
        explist = ExperimentListFactory.from_json_file(
            tmp_path / f"test_by_detector_{i:03d}.expt", check_format=False
        )
        refl = flex.reflection_table.from_file(
            tmp_path / f"test_by_detector_{i:03d}.refl"
        )

        for k in range(len(explist)):
            assert refl.experiment_identifiers()[k] == str(current_exp_id)
            current_exp_id += 2

    assert not tmp_path.joinpath("test_by_detector_002.expt").is_file()
    assert not tmp_path.joinpath("test_by_detector_002.refl").is_file()


@pytest.mark.parametrize("with_identifiers", ["True", "False"])
def test_combine_clustering(dials_data, tmp_path, with_identifiers):
    """Test with the clustering.use=True option.

    Need to use an integrated dataset for this option.
    """
    data_dir = dials_data("multi_crystal_proteinase_k", pathlib=True)

    input_range = [2, 3, 4, 5, 10]
    if with_identifiers:
        for n, i in enumerate(input_range):
            command = [
                "dials.assign_experiment_identifiers",
                data_dir / f"experiments_{i}.json",
                data_dir / f"reflections_{i}.pickle",
                f"output.experiments={n}.expt",
                f"output.reflections={n}.refl",
            ]
            procrunner.run(command, working_directory=tmp_path)

        phil_input = "\n".join(
            (
                f"  input.experiments={tmp_path / f'{i}.expt'}\n"
                + f"  input.reflections={tmp_path / f'{i}.refl'}"
            )
            for i in [0, 1, 2, 3, 4]
        )
    else:
        phil_input = "\n".join(
            (
                f"  input.experiments={data_dir}/experiments_{i}.json\n"
                + f"  input.reflections={data_dir}/reflections_{i}.pickle"
            )
        )

    tmp_path.joinpath("input.phil").write_text(phil_input)

    result = procrunner.run(
        [
            "dials.combine_experiments",
            tmp_path / "input.phil",
            "clustering.use=True",
            "threshold=5",
            "max_clusters=2",
        ],
        working_directory=tmp_path,
    )
    # this should create two clusters:
    #   combined_cluster_1 (2 expts)
    #   combined_cluster_2 (3 expts)

    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("combined_cluster2.refl").is_file()
    assert tmp_path.joinpath("combined_cluster2.expt").is_file()
    assert tmp_path.joinpath("combined_cluster1.refl").is_file()
    assert tmp_path.joinpath("combined_cluster1.expt").is_file()

    exps = load.experiment_list(tmp_path / "combined_cluster1.expt", check_format=False)
    assert len(exps) == 2
    refls = flex.reflection_table.from_file(tmp_path / "combined_cluster1.refl")
    assert list(set(refls["id"])) == [0, 1]
    exps = load.experiment_list(tmp_path / "combined_cluster2.expt", check_format=False)
    assert len(exps) == 3
    refls = flex.reflection_table.from_file(tmp_path / "combined_cluster2.refl")
    assert list(set(refls["id"])) == [0, 1, 2]


@pytest.fixture
def narrow_wedge_input_with_identifiers(dials_regression, tmpdir):
    """Make a fixture to avoid multiple runs of assign identifiers."""
    data_dir = os.path.join(
        dials_regression, "refinement_test_data", "multi_narrow_wedges"
    )
    input_range = [9, 11, 12, 31]
    for n, i in enumerate(input_range):
        command = [
            "dials.assign_experiment_identifiers",
            os.path.join(data_dir, "data/sweep_%03d/experiments.json" % i),
            os.path.join(data_dir, "data/sweep_%03d/reflections.pickle" % i),
            f"output.experiments={n}.expt",
            f"output.reflections={n}.refl",
        ]
        procrunner.run(command, working_directory=tmpdir)

    phil_input = "\n".join(
        (
            "  input.experiments=%s\n" % tmpdir.join("%s.expt" % i)
            + "  input.reflections=%s" % tmpdir.join("%s.refl" % i)
        )
        for i, _ in enumerate(input_range)
    )
    return phil_input


@pytest.mark.parametrize("min_refl", ["None", "100"])
@pytest.mark.parametrize("max_refl", ["None", "150"])
def test_min_max_reflections_per_experiment(
    dials_regression, tmp_path, min_refl, max_refl
):

    expected_results = {
        ("None", "None"): 10,
        ("None", "150"): 9,
        ("100", "None"): 6,
        ("100", "150"): 5,
    }

    data_dir = os.path.join(dials_regression, "refinement_test_data", "multi_stills")
    input_phil = (
        f" input.experiments={data_dir}/combined_experiments.json\n"
        + f" input.reflections={data_dir}/combined_reflections.pickle\n"
        + f" output.min_reflections_per_experiment={min_refl}\n"
        + f" output.max_reflections_per_experiment={max_refl}\n"
    ).format(data_dir, min_refl, max_refl)
    tmp_path.joinpath("input.phil").write_text(input_phil)

    result = procrunner.run(
        ["dials.combine_experiments", "input.phil"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr

    # load results
    exp = ExperimentListFactory.from_json_file(
        tmp_path / "combined.expt", check_format=False
    )

    assert len(exp) == expected_results[(min_refl, max_refl)]


@pytest.mark.parametrize("with_identifiers", ["True", "False"])
@pytest.mark.parametrize("method", ["random", "n_refl", "significance_filter"])
def test_combine_nsubset(
    dials_regression,
    tmpdir,
    with_identifiers,
    method,
    narrow_wedge_input_with_identifiers,
):
    """Test with the n_subset option."""

    if with_identifiers:
        phil_input = narrow_wedge_input_with_identifiers
    else:
        data_dir = os.path.join(
            dials_regression, "refinement_test_data", "multi_narrow_wedges"
        )
        input_range = [9, 11, 12, 31]
        phil_input = "\n".join(
            (
                "  input.experiments={0}/data/sweep_%03d/experiments.json\n"
                + "  input.reflections={0}/data/sweep_%03d/reflections.pickle"
            )
            % (i, i)
            for i in input_range
        ).format(data_dir)

    with open(tmpdir.join("input.phil").strpath, "w") as phil_file:
        phil_file.writelines(phil_input)

    result = procrunner.run(
        [
            "dials.combine_experiments",
            tmpdir.join("input.phil").strpath,
            "n_subset=3",
            f"n_subset_method={method}",
        ],
        working_directory=tmpdir,
    )
    assert not result.returncode and not result.stderr
    assert tmpdir.join("combined.refl").check()
    assert tmpdir.join("combined.expt").check()

    exps = load.experiment_list(
        tmpdir.join("combined.expt").strpath, check_format=False
    )
    assert len(exps) == 3
    refls = flex.reflection_table.from_file(tmpdir.join("combined.refl"))
    # Check that order are the same to ensure consistent for historical
    # use of ordered ids to match across datastructures
    assert list(exps.identifiers()) == list(refls.experiment_identifiers().values())
    assert len(set(refls["id"])) == 3
    assert list(set(refls["id"])) == [0, 1, 2]


def test_failed_tolerance_error(dials_regression, monkeypatch):
    """Test that we get a sensible error message on tolerance failures"""
    # Select some experiments to use for combining
    jsons = os.path.join(
        dials_regression,
        "refinement_test_data",
        "multi_narrow_wedges",
        "data",
        "sweep_{:03d}",
        "{}",
    )
    files = [
        jsons.format(2, "experiments.json"),
        jsons.format(2, "reflections.pickle"),
        jsons.format(3, "experiments.json"),
        jsons.format(3, "reflections.pickle"),
    ]

    # Use the combine script
    script = combine_experiments.Script()
    # Disable writing output
    monkeypatch.setattr(script, "_save_output", lambda *args: None)
    # Parse arguments and configure
    params, options = script.parser.parse_args(files)
    params.reference_from_experiment.beam = 0

    # Validate that these pass
    script.run_with_preparsed(params, options)

    # Now, alter the beam to check it doesn't pass
    exp_2 = params.input.experiments[1].data[0]
    exp_2.beam.set_wavelength(exp_2.beam.get_wavelength() * 2)

    with pytest.raises(SystemExit) as exc:
        script.run_with_preparsed(params, options)
    assert "Beam" in str(exc.value)
    print("Got (expected) error message:", exc.value)


def test_combine_imagesets(dials_data, tmp_path):
    data = dials_data("l_cysteine_dials_output", pathlib=True)
    args = [
        *sorted(data.glob("*_integrated.pickle")),
        *sorted(data.glob("*_integrated_experiments.json")),
        f"experiments_filename={tmp_path}/combined.expt",
        f"reflections_file={tmp_path}/combined.refl",
    ]
    combine_experiments.run([str(x) for x in args])

    comb = flex.reflection_table.from_file(tmp_path / "combined.refl")

    assert set(comb["imageset_id"]) == {0, 1, 2, 3}

    # Check that we have preserved unindexed reflections for all of these
    assert set(comb.select(comb["id"] == -1)["imageset_id"]) == {0, 1, 2, 3}



"""
Definitions of outlier rejection algorithms.

These algorithms use the Ih_table datastructures to perform calculations
in groups of symmetry equivalent reflections. Two functions are provided,
reject_outliers, to reject outlier and set flags given a reflection table
and experiment object, and determine_outlier_index_arrays, which takes an
Ih_table and returns flex.size_t index arrays of the outlier positions.
"""

from __future__ import annotations

import copy
import logging

import numpy as np

from dxtbx import flumpy
from scitbx.array_family import flex

from dials.algorithms.scaling.Ih_table import IhTable
from dials.util.normalisation import quasi_normalisation
from dials_scaling_ext import determine_outlier_indices, limit_outlier_weights

logger = logging.getLogger("dials")


def reject_outliers(reflection_table, experiment, method="standard", zmax=6.0):
    """
    Run an outlier algorithm on symmetry-equivalent intensities.

    This method runs an intensity-based outlier rejection algorithm, comparing
    the deviations from the weighted mean in groups of symmetry equivalent
    reflections. The outliers are determined and the outlier_in_scaling flag
    is set in the reflection table.

    The values intensity and variance must be set in the reflection table;
    these should be corrected but unscaled values, as an inverse_scale_factor
    will be applied during outlier rejection if this is present in the reflection
    table. The reflection table should also be prefiltered (e.g. not-integrated
    reflections should not be present) as no further filtering is done on the
    input table.

    Args:
        reflection_table: A reflection table.
        experiment: A single experiment object.
        method (str): Name (alias) of outlier rejection algorithm to use.
        zmax (float): Normalised deviation threshold for classifying an outlier.

    Returns:
        reflection_table: The input table with the outlier_in_scaling flag set.
    """
    assert "intensity" in reflection_table, "reflection table has no 'intensity' column"
    assert "variance" in reflection_table, "reflection table has no 'variance' column"

    if "inverse_scale_factor" not in reflection_table:
        reflection_table["inverse_scale_factor"] = flex.double(
            reflection_table.size(), 1.0
        )

    Ih_table = IhTable(
        [reflection_table], experiment.crystal.get_space_group(), nblocks=1
    )
    outlier_indices = determine_outlier_index_arrays(
        Ih_table, method=method, zmax=zmax
    )[0]

    # Unset any existing outlier flags before setting the new ones
    reflection_table.unset_flags(
        reflection_table.get_flags(reflection_table.flags.outlier_in_scaling),
        reflection_table.flags.outlier_in_scaling,
    )
    reflection_table.set_flags(
        flumpy.from_numpy(outlier_indices), reflection_table.flags.outlier_in_scaling
    )

    return reflection_table


def determine_Esq_outlier_index_arrays(Ih_table, experiment, emax=10.0):
    # first calculate normalised intensities and set in the Ih_table.
    intensities = Ih_table.as_miller_array(experiment.crystal.get_unit_cell())
    normalised_intensities = quasi_normalisation(intensities)

    sel = normalised_intensities.data() > (emax**2)
    n_e2_outliers = sel.count(True)
    if n_e2_outliers:
        logger.info(
            f"{n_e2_outliers} outliers identified from normalised intensity analysis (E\xb2 > {(emax ** 2)})"
        )
    outlier_indices = (
        Ih_table.Ih_table_blocks[0]
        .Ih_table["loc_indices"]
        .iloc[flumpy.to_numpy(sel)]
        .to_numpy()
    )
    if Ih_table.n_datasets == 1:
        return [outlier_indices]
    datasets = (
        Ih_table.Ih_table_blocks[0]
        .Ih_table["dataset_id"]
        .iloc[flumpy.to_numpy(sel)]
        .to_numpy()
    )
    final_outlier_arrays = []
    for i in range(Ih_table.n_datasets):
        final_outlier_arrays.append(outlier_indices[datasets == i])
    return final_outlier_arrays


def determine_outlier_index_arrays(Ih_table, method="standard", zmax=6.0, target=None):
    """
    Run an outlier algorithm and return the outlier indices.

    Args:
        Ih_table: A dials.algorithms.scaling.Ih_table.IhTable.
        method (str): Name (alias) of outlier rejection algorithm to use. If
            method=target, then the optional argument target must also
            be specified. Implemented methods; standard, simple, target.
        zmax (float): Normalised deviation threshold for classifying an outlier.
        target (Optional[IhTable]): An IhTable to use to obtain target Ih for
            outlier rejectiob, if method=target.

    Returns:
        outlier_index_arrays (list): A list of flex.size_t arrays, with one
            array per dataset that was used to create the Ih_table. Importantly,
            the indices are the indices of the reflections in the initial
            reflection table used to create the Ih_table, not the indices of the
            data in the Ih_table.

    Raises:
        ValueError: if an invalid choice is made for the method.
    """
    outlier_rej = None
    if method == "standard":
        outlier_rej = NormDevOutlierRejection(Ih_table, zmax)
    elif method == "simple":
        outlier_rej = SimpleNormDevOutlierRejection(Ih_table, zmax)
    elif method == "target":
        assert target is not None
        outlier_rej = TargetedOutlierRejection(Ih_table, zmax, target)
    elif method is not None:
        raise ValueError(f"Invalid choice of outlier rejection method: {method}")
    if not outlier_rej:
        return [flex.size_t([]) for _ in range(Ih_table.n_datasets)]
    outlier_rej.run()
    outlier_index_arrays = outlier_rej.final_outlier_arrays
    if Ih_table.n_datasets > 1:
        msg = (
            "Combined outlier rejection has been performed across multiple datasets, \n"
        )
    else:
        msg = "A round of outlier rejection has been performed, \n"
    n_outliers = sum(len(i) for i in outlier_index_arrays)
    msg += f"{n_outliers} outliers have been identified. \n"
    logger.info(msg)
    return outlier_index_arrays


class OutlierRejectionBase:
    """
    Base class for outlier rejection algorithms using an IhTable datastructure.

    Subclasses must implement the _do_outlier_rejection method, which must
    add the indices of outliers to the _outlier_indices attribute. The algorithms
    are run upon initialisation and result in the population of the
    :obj:`final_outlier_arrays`.

    Attributes:
        final_outlier_arrays (:obj:`list`): A list of flex.size_t arrays of outlier
            indices w.r.t. the order of the initial reflection tables used to
            create the Ih_table.
    """

    def __init__(self, Ih_table, zmax):
        """Set up and run the outlier rejection algorithm."""
        assert (
            Ih_table.n_work_blocks == 1
        ), """
Outlier rejection algorithms require an Ih_table with nblocks = 1"""
        # Note: could be possible to code for nblocks > 1
        self._Ih_table_block = Ih_table.blocked_data_list[0]
        self._n_datasets = Ih_table.n_datasets
        self._block_selections = Ih_table.blocked_selection_list[0]
        self._datasets = np.array([], dtype=np.uint64).reshape((0,))  # flex.int([])
        self._zmax = zmax
        self._outlier_indices = np.array([], dtype=np.uint64).reshape(
            (0,)
        )  # flex.size_t([])
        self.final_outlier_arrays = None

    def run(self):
        """Run the outlier rejection algorithm, implemented by a subclass."""
        self._do_outlier_rejection()
        self.final_outlier_arrays = self._determine_outlier_indices()

    def _determine_outlier_indices(self):
        """
        Determine outlier indices with respect to the input reflection tables.

        Transform the outlier indices w.r.t the Ih_table, determined during the
        algorithm, to outlier indices w.r.t the initial reflection tables used
        to create the Ih_table, separated by reflection table.

        Returns:
            final_outlier_arrays (:obj:`list`): A list of flex.size_t arrays of
                outlier indices w.r.t. the order of the data in the initial
                reflection tables used to create the Ih_table.
        """
        if self._n_datasets == 1:
            return [self._outlier_indices]
        final_outlier_arrays = []
        for i in range(self._n_datasets):
            final_outlier_arrays.append(self._outlier_indices[self._datasets == i])
        return final_outlier_arrays

    def _do_outlier_rejection(self):
        """Add indices (w.r.t. the Ih_table data) to self._outlier_indices."""
        raise NotImplementedError()


class TargetedOutlierRejection(OutlierRejectionBase):
    """Implementation of an outlier rejection algorithm against a target.

    This algorithm requires a target Ih_table in addition to an Ih_table
    for the dataset under investigation. Normalised deviations are
    calculated from the intensity values in the target table.
    """

    def __init__(self, Ih_table, zmax, target):
        """Set a target Ih_table and run the outlier rejection."""
        assert (
            target.n_work_blocks == 1
        ), """
Targeted outlier rejection requires a target Ih_table with nblocks = 1"""
        self._target_Ih_table_block = target.blocked_data_list[0]
        self._target_Ih_table_block.calc_Ih()
        super().__init__(Ih_table, zmax)

    def _do_outlier_rejection(self):
        """Add indices (w.r.t. the Ih_table data) to self._outlier_indices."""
        Ih_table = self._Ih_table_block
        target = self._target_Ih_table_block
        target_asu_Ih_dict = dict(
            zip(target.asu_miller_index, zip(target.Ih_values, target.variances))
        )
        target_Ih_value = np.zeros(Ih_table.size)
        target_Ih_sigmasq = np.zeros(Ih_table.size)
        for j, miller_idx in enumerate(Ih_table.asu_miller_index):
            try:
                vals = target_asu_Ih_dict[miller_idx]
            except KeyError:
                pass
            else:
                target_Ih_value[j] = vals[0]
                target_Ih_sigmasq[j] = vals[1]

        nz_sel = target_Ih_value != 0.0
        target_Ih_value = target_Ih_value[nz_sel]
        target_Ih_sigmasq = target_Ih_sigmasq[nz_sel]
        Ih_table = Ih_table.select(nz_sel)
        norm_dev = (
            Ih_table.intensities - (Ih_table.inverse_scale_factors * target_Ih_value)
        ) / (
            np.sqrt(
                Ih_table.variances
                + (np.square(Ih_table.inverse_scale_factors) * target_Ih_sigmasq)
            )
        )
        outliers_sel = np.abs(norm_dev) > self._zmax
        outliers_isel = np.nonzero(nz_sel)[0][outliers_sel]

        outliers = np.full(self._Ih_table_block.size, False)
        outliers[outliers_isel] = True

        self._outlier_indices = np.concatenate(
            [
                self._outlier_indices,
                self._Ih_table_block.Ih_table["loc_indices"].iloc[outliers],
            ]
        )
        self._datasets = np.concatenate(
            [
                self._datasets,
                self._Ih_table_block.Ih_table["dataset_id"].iloc[outliers],
            ]
        )


class SimpleNormDevOutlierRejection(OutlierRejectionBase):
    """Algorithm using normalised deviations from the weighted intensity means.

    In this case, the weighted mean is calculated from all reflections in
    the symmetry group excluding the test reflection.
    """

    def __init__(self, Ih_table, zmax):
        super().__init__(Ih_table, zmax)
        self.weights = flumpy.to_numpy(
            limit_outlier_weights(
                copy.deepcopy(self._Ih_table_block.weights),
                self._Ih_table_block.h_index_matrix,
            )
        )

    def _do_outlier_rejection(self):
        """Add indices (w.r.t. the Ih_table data) to self._outlier_indices."""
        Ih_table = self._Ih_table_block
        intensity = Ih_table.intensities
        g = Ih_table.inverse_scale_factors
        w = self.weights
        wgIsum = Ih_table.sum_in_groups(w * g * intensity, output="per_refl")
        wg2sum = Ih_table.sum_in_groups(w * g * g, output="per_refl")

        # guard against zero division errors - can happen due to rounding errors
        # or bad data giving g values are very small
        zero_sel = wg2sum == 0.0
        # set as one for now, then mark as outlier below. This will only affect if
        # g is near zero, if w is zero then throw an assertionerror.
        wg2sum[zero_sel] = 1.0

        assert np.all(w > 0)  # guard against division by zero
        norm_dev = (intensity - (g * wgIsum / wg2sum)) / (
            np.sqrt((1.0 / w) + np.square(g / wg2sum))
        )
        norm_dev[zero_sel] = 1000  # to trigger rejection
        outliers = np.abs(norm_dev) > self._zmax

        self._outlier_indices = np.concatenate(
            [
                self._outlier_indices,
                self._Ih_table_block.Ih_table["loc_indices"].iloc[outliers].to_numpy(),
            ]
        )
        self._datasets = np.concatenate(
            [
                self._datasets,
                self._Ih_table_block.Ih_table["dataset_id"].iloc[outliers].to_numpy(),
            ]
        )


class NormDevOutlierRejection(OutlierRejectionBase):
    """Algorithm using normalised deviations from the weighted intensity means.

    In this case, the weighted mean is calculated from all reflections in
    the symmetry group excluding the test reflection.
    """

    def __init__(self, Ih_table, zmax):
        super().__init__(Ih_table, zmax)
        self.weights = flumpy.to_numpy(
            limit_outlier_weights(
                copy.deepcopy(self._Ih_table_block.weights),
                self._Ih_table_block.h_index_matrix,
            )
        )

    def _do_outlier_rejection(self):
        """Add indices (w.r.t. the Ih_table data) to self._outlier_indices."""
        self._round_of_outlier_rejection()
        n_outliers = self._outlier_indices.size
        n_new_outliers = n_outliers
        while n_new_outliers:
            self._round_of_outlier_rejection()
            n_new_outliers = self._outlier_indices.size - n_outliers
            n_outliers = self._outlier_indices.size

    def _round_of_outlier_rejection(self):
        """
        Calculate normal deviations from the data in the Ih_table.
        """
        Ih_table = self._Ih_table_block
        intensity = Ih_table.intensities
        g = Ih_table.inverse_scale_factors
        w = self.weights
        wgIsum = Ih_table.sum_in_groups(w * g * intensity, output="per_refl")
        wg2sum = Ih_table.sum_in_groups(w * g * g, output="per_refl")
        wgIsum_others = wgIsum - (w * g * intensity)
        wg2sum_others = wg2sum - (w * g * g)
        # Now do the rejection analysis if n_in_group > 2
        nh = Ih_table.calc_nh()
        sel = nh > 2
        wg2sum_others_sel = wg2sum_others[sel]
        wgIsum_others_sel = wgIsum_others[sel]

        # guard against zero division errors - can happen due to rounding errors
        # or bad data giving g values are very small
        zero_sel = wg2sum_others_sel == 0.0
        # set as one for now, then mark as outlier below. This will only affect if
        # g is near zero, if w is zero then throw an assertionerror.
        wg2sum_others_sel[zero_sel] = 1.0
        g_sel = g[sel]
        I_sel = intensity[sel]
        w_sel = w[sel]

        assert np.all(w_sel > 0)  # guard against division by zero
        norm_dev = (I_sel - (g_sel * wgIsum_others_sel / wg2sum_others_sel)) / (
            np.sqrt((1.0 / w_sel) + (np.square(g_sel) / wg2sum_others_sel))
        )
        norm_dev[zero_sel] = 1000  # to trigger rejection
        z_score = np.abs(norm_dev)
        # Want an array same size as Ih table.
        all_z_scores = np.zeros(Ih_table.size)
        # all_z_scores.set_selected(sel.iselection(), z_score)
        all_z_scores[sel] = z_score
        outlier_indices, other_potential_outliers = determine_outlier_indices(
            Ih_table.h_index_matrix, flumpy.from_numpy(all_z_scores), self._zmax
        )
        sel = np.full(Ih_table.size, False, dtype=bool)
        outlier_indices = flumpy.to_numpy(outlier_indices)
        sel[outlier_indices] = True
        lsel = self._Ih_table_block.Ih_table["loc_indices"].iloc[sel].to_numpy()
        dsel = self._Ih_table_block.Ih_table["dataset_id"].iloc[sel].to_numpy()
        self._outlier_indices = np.concatenate(
            [
                self._outlier_indices,
                lsel,
            ]
        )
        self._datasets = np.concatenate(
            [
                self._datasets,
                dsel,
            ]
        )

        sel = np.full(Ih_table.size, False, dtype=bool)
        sel[flumpy.to_numpy(other_potential_outliers)] = True
        self._Ih_table_block = self._Ih_table_block.select(sel)
        self.weights = self.weights[sel]



from __future__ import annotations

import boost_adaptbx.boost.python

ext = boost_adaptbx.boost.python.import_ext("dials_model_data_ext")

from dials_model_data_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "AdjacencyList",
    "AdjacentVerticesIter",
    "Centroid",
    "CentroidData",
    "EdgeDescriptor",
    "ImageDouble",
    "ImageInt",
    "ImageVolume",
    "Intensity",
    "IntensityData",
    "MultiPanelImageVolume",
    "Observation",
    "PixelList",
    "PixelListLabeller",
    "PositionData",
    "Prediction",
    "Ray",
    "Shoebox",
    "make_image",
)
